# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Spyder
======

The Scientific Python Development Environment

Spyder is a powerful scientific environment written in Python, for Python,
and designed by and for scientists, engineers and data analysts.

It features a unique combination of the advanced editing, analysis, debugging
and profiling functionality of a comprehensive development tool with the data
exploration, interactive execution, deep inspection and beautiful visualization
capabilities of a scientific package.
"""

from __future__ import print_function

# Standard library imports
from distutils.command.install_data import install_data
import io
import os
import os.path as osp
import subprocess
import sys

# Third party imports
from setuptools import setup, find_packages
from setuptools.command.install import install


# =============================================================================
# Minimal Python version sanity check
# Taken from the notebook setup.py -- Modified BSD License
# =============================================================================
v = sys.version_info
if v[0] >= 3 and v[:2] < (3, 8):
    error = "ERROR: Spyder requires Python version 3.8 and above."
    print(error, file=sys.stderr)
    sys.exit(1)


# =============================================================================
# Constants
# =============================================================================
NAME = 'spyder'
LIBNAME = 'spyder'
WINDOWS_INSTALLER_NAME = os.environ.get('EXE_NAME')

from spyder import __version__, __website_url__  #analysis:ignore


# =============================================================================
# Auxiliary functions
# =============================================================================
def get_package_data(name, extlist):
    """
    Return data files for package *name* with extensions in *extlist*.
    """
    flist = []
    # Workaround to replace os.path.relpath (not available until Python 2.6):
    offset = len(name)+len(os.pathsep)
    for dirpath, _dirnames, filenames in os.walk(name):
        if 'tests' not in dirpath:
            for fname in filenames:
                if (not fname.startswith('.') and
                        osp.splitext(fname)[1] in extlist):
                    flist.append(osp.join(dirpath, fname)[offset:])
    return flist


def get_subpackages(name):
    """
    Return subpackages of package *name*.
    """
    splist = []
    for dirpath, _dirnames, _filenames in os.walk(name):
        if 'tests' not in dirpath:
            if osp.isfile(osp.join(dirpath, '__init__.py')):
                splist.append(".".join(dirpath.split(os.sep)))

    return splist


def get_data_files():
    """
    Return data_files in a platform dependent manner.
    """
    if sys.platform.startswith('linux'):
        data_files = [('share/applications', ['scripts/spyder.desktop']),
                      ('share/icons', ['img_src/spyder.png']),
                      ('share/metainfo',
                       ['scripts/org.spyder_ide.spyder.appdata.xml'])]
    elif os.name == 'nt':
        data_files = [('scripts', ['img_src/spyder.ico'])]
    else:
        data_files = []

    return data_files


def get_packages():
    """
    Return package list.
    """
    packages = get_subpackages(LIBNAME)
    return packages


def get_qt_requirements(qt_requirements, default='pyqt5'):
    """
    Return a list of requirements for the Qt binding according to the
    environment variable SPYDER_QT_BINDING. If this variable is not set
    or has an unsupported value it defaults to 'pyqt5'.

    Parameters
    ----------
    qt_requirements : dict
        A dictionary whose keys are supported Qt bindings and whose values are
        lists of required packages to install for each binding.
    default : str
        Default Qt binding to use if the environment variable is not set.
        Defaults to 'pyqt5'.

    Raises
    ------
    ValueError
        If the environment variable SPYDER_QT_BINDING has an unsupported value.

    Returns
    -------
    install_requires : list
        A list of required packages to install for the given Qt binding.
    """
    install_requires = []

    # Check if a Qt binding is set in the environment and normalizes
    env_qt_binding = os.environ.get('SPYDER_QT_BINDING', default)
    env_qt_binding = env_qt_binding.lower()
    install_requires = qt_requirements.get(env_qt_binding, None)

    if install_requires is None:
        raise ValueError(
            f"Unsupported Qt binding: {env_qt_binding}. "
            f"Supported: "  + ", ".join(qt_requirements.keys())
        )

    return install_requires


# =============================================================================
# Make Linux detect Spyder desktop file (will not work with wheels)
# =============================================================================
class CustomInstallData(install_data):

    def run(self):
        install_data.run(self)
        if sys.platform.startswith('linux'):
            try:
                subprocess.call(['update-desktop-database'])
            except:
                print("ERROR: unable to update desktop database",
                      file=sys.stderr)


CMDCLASS = {'install_data': CustomInstallData}


# =============================================================================
# Main scripts
# =============================================================================
# NOTE: the '[...]_win_post_install.py' script is installed even on non-Windows
# platforms due to a bug in pip installation process
# See spyder-ide/spyder#1158.
SCRIPTS = ['%s_win_post_install.py' % NAME]

SCRIPTS.append('spyder')

if os.name == 'nt':
    SCRIPTS += ['spyder.bat']


# =============================================================================
# Files added to the package
# =============================================================================
EXTLIST = ['.pot', '.po', '.mo', '.svg', '.png', '.css', '.html', '.js',
           '.ini', '.txt', '.qss', '.ttf', '.json', '.rst', '.bloom',
           '.ico', '.gif', '.mp3', '.ogg', '.sfd', '.bat', '.sh']


# =============================================================================
# Use Readme for long description
# =============================================================================
with io.open('README.md', encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()


#==============================================================================
# Setup arguments
#==============================================================================
setup_args = dict(
    name=NAME,
    version=__version__,
    description='The Scientific Python Development Environment',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    download_url=__website_url__ + "#fh5co-download",
    author="The Spyder Project Contributors",
    author_email="spyder.python@gmail.com",
    url=__website_url__,
    license='MIT',
    keywords='PyQt5 editor console widgets IDE science data analysis IPython',
    platforms=["Windows", "Linux", "Mac OS-X"],
    packages=get_packages(),
    package_data={LIBNAME: get_package_data(LIBNAME, EXTLIST)},
    scripts=[osp.join('scripts', fname) for fname in SCRIPTS],
    data_files=get_data_files(),
    python_requires='>=3.8',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Widget Sets',
    ],
    cmdclass=CMDCLASS,
)

# Qt bindings requirements
qt_requirements = {
    'pyqt5': [
        'pyqt5>=5.15,<5.16',
        'pyqt5-sip<12.16; python_version=="3.8"',
        'pyqtwebengine>=5.15,<5.16',
        'qtconsole>=5.6.1,<5.7.0',
    ],
    'pyqt6': [
        'pyqt6>=6.5,<7',
        'pyqt6-webengine>=6.5,<7',
        'qtconsole>=5.6.1,<5.7.0',
    ],
    'conda-forge': [
        'qtconsole>=5.6.1,<5.7.0',
    ]
}

# Get the proper requirements for the selected Qt binding
install_requires = get_qt_requirements(qt_requirements, default='pyqt5')

install_requires += [
    'aiohttp>=3.9.3',
    'applaunchservices>=0.3.0;platform_system=="Darwin"',
    'asyncssh>=2.14.0,<3.0.0',
    'atomicwrites>=1.2.0',
    'chardet>=2.0.0',
    'cloudpickle>=0.5.0',
    'cookiecutter>=1.6.0',
    'diff-match-patch>=20181111',
    # While this is only required for python <3.10, it is safe enough to
    # install in all cases and helps the tests to pass.
    'importlib-metadata>=4.6.0',
    'intervaltree>=3.0.2',
    'ipython>=8.12.2,<8.13.0; python_version=="3.8"',
    'ipython>=8.13.0,<9.0.0,!=8.17.1; python_version>"3.8"',
    'jedi>=0.17.2,<0.20.0',
    'jellyfish>=0.7',
    'jsonschema>=3.2.0',
    'keyring>=17.0.0',
    'nbconvert>=4.0',
    'numpydoc>=0.6.0',
    'parso>=0.7.0,<0.9.0',
    'pexpect>=4.4.0',
    'pickleshare>=0.4',
    'psutil>=5.3',
    'pygithub>=2.3.0',
    'pygments>=2.0',
    'pylint>=3.1,<4',
    'pylint-venv>=3.0.2',
    'pyls-spyder>=0.4.0',
    'python-lsp-black>=2.0.0,<3.0.0',
    'python-lsp-server[all]>=1.12.0,<1.13.0',
    'pyuca>=1.2',
    'pyxdg>=0.26;platform_system=="Linux"',
    'pyzmq>=24.0.0',
    'qdarkstyle>=3.2.0,<3.3.0',
    'qstylizer>=0.2.2',
    'qtawesome>=1.4.0,<1.5.0',
    'qtpy>=2.4.0',
    'rtree>=0.9.7',
    'setuptools>=49.6.0',
    'sphinx>=0.6.6',
    'spyder-kernels>=3.0.3,<3.1.0',
    'superqt>=0.6.2,<1.0.0',
    'textdistance>=4.2.0',
    'three-merge>=0.1.1',
    'watchdog>=0.10.3',
    'yarl>=1.9.4',
]

# Loosen constraints to ensure dev versions still work
if 'dev' in __version__:
    reqs_to_loosen = {
        'python-lsp-server[all]',
        'qtconsole',
        'qtconsole-base',
        'spyder-kernels',
    }
    install_requires = [req for req in install_requires
                        if req.split(">")[0] not in reqs_to_loosen]

    install_requires.append('python-lsp-server[all]>=1.12.0,<1.14.0')
    install_requires.append('qtconsole>=5.5.1,<5.7.0')

extras_require = {
    'test:platform_system == "Windows"': ['pywin32'],
    'test': [
        'coverage',
        'cython',
        'flaky',
        'matplotlib',
        'pandas',
        'pillow',
        'pytest<8.0',
        'pytest-cov',
        'pytest-lazy-fixture',
        'pytest-mock',
        'pytest-order',
        'pytest-qt',
        'pytest-timeout',
        'pyyaml',
        'scipy',
        'sympy',
    ],
}


spyder_plugins_entry_points = [
    'appearance = spyder.plugins.appearance.plugin:Appearance',
    'application = spyder.plugins.application.plugin:Application',
    'completions = spyder.plugins.completion.plugin:CompletionPlugin',
    'debugger = spyder.plugins.debugger.plugin:Debugger',
    'editor = spyder.plugins.editor.plugin:Editor',
    'explorer = spyder.plugins.explorer.plugin:Explorer',
    'external_terminal = spyder.plugins.externalterminal.plugin:ExternalTerminal',
    'find_in_files = spyder.plugins.findinfiles.plugin:FindInFiles',
    'help = spyder.plugins.help.plugin:Help',
    'historylog = spyder.plugins.history.plugin:HistoryLog',
    'internal_console = spyder.plugins.console.plugin:Console',
    'ipython_console = spyder.plugins.ipythonconsole.plugin:IPythonConsole',
    'layout = spyder.plugins.layout.plugin:Layout',
    'main_interpreter = spyder.plugins.maininterpreter.plugin:MainInterpreter',
    'mainmenu = spyder.plugins.mainmenu.plugin:MainMenu',
    'onlinehelp = spyder.plugins.onlinehelp.plugin:OnlineHelp',
    'outline_explorer = spyder.plugins.outlineexplorer.plugin:OutlineExplorer',
    'plots = spyder.plugins.plots.plugin:Plots',
    'preferences = spyder.plugins.preferences.plugin:Preferences',
    'profiler = spyder.plugins.profiler.plugin:Profiler',
    'project_explorer = spyder.plugins.projects.plugin:Projects',
    'pylint = spyder.plugins.pylint.plugin:Pylint',
    'pythonpath_manager = spyder.plugins.pythonpath.plugin:PythonpathManager',
    'remoteclient = spyder.plugins.remoteclient.plugin:RemoteClient',
    'run = spyder.plugins.run.plugin:Run',
    'shortcuts = spyder.plugins.shortcuts.plugin:Shortcuts',
    'statusbar = spyder.plugins.statusbar.plugin:StatusBar',
    'switcher = spyder.plugins.switcher.plugin:Switcher',
    'toolbar = spyder.plugins.toolbar.plugin:Toolbar',
    'tours = spyder.plugins.tours.plugin:Tours',
    'update_manager = spyder.plugins.updatemanager.plugin:UpdateManager',
    'variable_explorer = spyder.plugins.variableexplorer.plugin:VariableExplorer',
    'workingdir = spyder.plugins.workingdirectory.plugin:WorkingDirectory',
]

spyder_completions_entry_points = [
    ('fallback = spyder.plugins.completion.providers.fallback.provider:'
     'FallbackProvider'),
    ('snippets = spyder.plugins.completion.providers.snippets.provider:'
     'SnippetsProvider'),
    ('lsp = spyder.plugins.completion.providers.languageserver.provider:'
     'LanguageServerProvider'),
]


setup_args['install_requires'] = install_requires
setup_args['extras_require'] = extras_require
setup_args['entry_points'] = {
    'gui_scripts': [
            'spyder = spyder.app.start:main'
    ],
    'spyder.plugins': spyder_plugins_entry_points,
    'spyder.completions': spyder_completions_entry_points
}
setup_args.pop('scripts', None)


# =============================================================================
# Main setup
# =============================================================================
setup(**setup_args)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2021, Spyder Bot
#
# Licensed under the terms of the MIT license
# ----------------------------------------------------------------------------

from setuptools import find_packages
from setuptools import setup

from spyder_boilerplate import __version__


setup(
    # See: https://setuptools.readthedocs.io/en/latest/setuptools.html
    name="spyder-boilerplate",
    version=__version__,
    author="Spyder Bot",
    author_email="spyder.python@gmail.com",
    description="Plugin that registers a programmatic custom layout",
    license="MIT license",
    python_requires='>= 3.8',
    install_requires=[
        "qtpy",
        "qtawesome",
        "spyder>=6",
    ],
    packages=find_packages(),
    entry_points={
        "spyder.plugins": [
            "spyder_boilerplate = spyder_boilerplate.spyder.plugin:SpyderBoilerplate"
        ],
    },
    classifiers=[
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
    ],
)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# ----------------------------------------------------------------------------

"""Tests for the plugin."""

# Standard library imports
import os
import sys

# Third party imports
import pytest

# Local imports
from spyder.config.base import running_in_ci
from spyder.config.utils import is_anaconda
from spyder.plugins.maininterpreter.plugin import MainInterpreter


@pytest.fixture
def maininterpreter(qtbot):
    """Set up PathManager."""
    plugin = MainInterpreter(None)
    qtbot.addWidget(plugin.get_container())
    return plugin


@pytest.mark.skipif(not is_anaconda(), reason="Requires conda to be installed")
@pytest.mark.skipif(not running_in_ci(), reason="Only meant for CIs")
def test_conda_interpreters(maininterpreter, qtbot):
    """Test info from conda interpreters."""
    container = maininterpreter.get_container()
    container._interpreter = ''

    name_base = 'Conda: base'
    name_test = 'Conda: jedi-test-env'

    # Wait until envs are computed
    qtbot.wait(4000)

    # Update to the base conda environment
    path_base, version = container.envs[name_base]
    expected = 'Conda: base ({})'.format(version)
    assert expected == container._get_env_info(path_base)

    # Update to the foo conda environment
    path_foo, version = container.envs[name_test]
    expected = 'Conda: jedi-test-env ({})'.format(version)
    assert expected == container._get_env_info(path_foo)


def test_pyenv_interpreters(maininterpreter, qtbot):
    """Test info from pyenv interpreters."""
    container = maininterpreter.get_container()

    version = 'Python 3.6.6'
    name = 'pyenv: test'
    interpreter = os.sep.join(['some-other', 'bin', 'python'])
    container.envs = {name: (interpreter, version)}
    container.path_to_env = {interpreter: name}
    assert 'pyenv: test (Python 3.6.6)' == container._get_env_info(interpreter)


@pytest.mark.skipif(sys.platform != 'darwin', reason="Only valid on Mac")
def test_internal_interpreter(maininterpreter, qtbot, mocker):
    """Test info from internal interpreter."""
    container = maininterpreter.get_container()

    interpreter = os.sep.join(['Spyder.app', 'Contents', 'MacOS', 'Python'])
    name = 'system:'
    version = 'Python 3.6.6'
    container.envs = {name: (interpreter, version)}
    container.path_to_env = {interpreter: name}
    assert 'system: (Python 3.6.6)' == container._get_env_info(interpreter)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

from unittest.mock import MagicMock

from qtpy.QtWidgets import QMainWindow
import pytest

from spyder.config.manager import CONF
from spyder.plugins.toolbar.api import ApplicationToolbars
from spyder.plugins.toolbar.plugin import Toolbar
from spyder.plugins.workingdirectory.plugin import WorkingDirectory


class MainWindow(QMainWindow):

    _cli_options = MagicMock()

    def get_plugin(self, name, error=True):
        return MagicMock()


@pytest.fixture
def toolbar(qtbot):
    # Create toolbar plugin
    main_window = MainWindow()
    toolbar = Toolbar(main_window, configuration=CONF)
    toolbar.on_initialize()

    # Add working directory toolbar
    cwd = WorkingDirectory(main_window, None)
    cwd.on_initialize()
    toolbar.add_application_toolbar(cwd.get_container().toolbar)

    # Add buttons to the other toolbars
    actions = [
        (ApplicationToolbars.File, "filenew"),
        (ApplicationToolbars.Run, "run"),
        (ApplicationToolbars.Debug, "debug"),
        (ApplicationToolbars.Main, "python"),
    ]

    for i, item in enumerate(actions):
        toolbar.create_action(
            f"action_{i}",
            text="Action {i}",
            icon=toolbar.create_icon(item[1]),
            triggered=lambda: True,
        )

        toolbar.add_item_to_application_toolbar(
            toolbar.get_action(f"action_{i}"),
            toolbar_id=item[0],
        )

    # Show toolbars and main window
    toolbar.on_mainwindow_visible()
    toolbar.toggle_lock(True)
    main_window.show()
    return toolbar


def test_default_order(toolbar):
    """Check toolbars are displayed in their default order."""
    expected_order = [
        ApplicationToolbars.File,
        ApplicationToolbars.Run,
        ApplicationToolbars.Debug,
        ApplicationToolbars.Main,
        ApplicationToolbars.WorkingDirectory,
    ]

    # Check the horizontal order obtained the displayed toolbar doesn't change
    # after sorting it.
    current_order = [
        toolbar.get_application_toolbar(id_).x()
        for id_ in expected_order
    ]
    assert current_order == sorted(current_order)


def test_restore_toolbars_order(toolbar, qtbot):
    """
    Check that if a toolbar is moved by users to a different position, then the
    new order is restored during the next startup.
    """
    new_order = [
        ApplicationToolbars.Debug,
        ApplicationToolbars.File,
        ApplicationToolbars.Run,
        ApplicationToolbars.Main,
    ]

    # Select two toolbars to change their order
    file_toolbar = toolbar.get_application_toolbar(ApplicationToolbars.File)
    debug_toolbar = toolbar.get_application_toolbar(ApplicationToolbars.Debug)

    # Move the debug toolbar to be before the file one
    toolbar.get_main().insertToolBar(file_toolbar, debug_toolbar)
    qtbot.wait(500)
    assert debug_toolbar.x() < file_toolbar.x()

    # Close toolbar plugin (which is called when the main window is closed)
    toolbar.on_close(False)

    # Check the list of available toolbars was saved to our config system
    assert toolbar.get_conf("last_toolbars")

    # Reload toolbars as if a new Spyder session was started
    toolbar.on_mainwindow_visible()
    qtbot.wait(500)

    # Check the order of displayed toolbars is the expected one
    current_order = [
        toolbar.get_application_toolbar(id_).pos().x()
        for id_ in new_order
    ]
    assert current_order == sorted(current_order)


def test_reorder_toolbars(toolbar, qtbot):
    """
    Check that if a toolbar is removed, then toolbars are reordered the next
    time with the working directory being shown to the right.
    """
    # Select the toolbars we are going to move
    cwd_toolbar = toolbar.get_application_toolbar(
        ApplicationToolbars.WorkingDirectory
    )
    debug_toolbar = toolbar.get_application_toolbar(ApplicationToolbars.Debug)

    # Move the working directory toolbar to be before the debug one
    toolbar.get_main().insertToolBar(debug_toolbar, cwd_toolbar)
    qtbot.wait(500)
    assert cwd_toolbar.x() < debug_toolbar.x()

    # Close toolbar plugin (which is called when the main window is closed)
    toolbar.on_close(False)
    qtbot.wait(500)

    # Simulate that a toolbar was present in the last sesstion and then removed
    last_toolbars = toolbar.get_conf("last_toolbars")
    toolbar.set_conf("last_toolbars", last_toolbars + ["foo_toolbar"])

    # Reload toolbars as if a new Spyder session was started
    toolbar.on_mainwindow_visible()

    # Check the working directory toolbar is to the right of the rest
    current_order = [app_toolbar.x() for app_toolbar in toolbar.toolbarslist]
    assert all([cwd_toolbar.x() >= pos for pos in current_order])



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tests for the Spyder `help` plugn, `help.py`.
"""

# Standard library imports
from unittest.mock import Mock

# Third party imports
from qtpy.QtWidgets import QMainWindow
from qtpy.QtWebEngineWidgets import WEBENGINE
import pytest
from flaky import flaky

# Local imports
from spyder.plugins.help.plugin import Help


# =============================================================================
# Fixtures
# =============================================================================
@pytest.fixture
def help_plugin(qtbot):
    """Help plugin fixture"""

    class MainMock(QMainWindow):
        def __getattr__(self, attr):
            if attr == 'ipyconsole' or attr == 'editor':
                return None
            else:
                return Mock()

    window = MainMock()
    help_plugin = Help(parent=window, configuration=None)
    help_widget = help_plugin.get_widget()
    webview = help_widget.rich_text.webview._webview

    window.setCentralWidget(help_widget)

    if WEBENGINE:
        help_widget._webpage = webview.page()
    else:
        help_widget._webpage = webview.page().mainFrame()

    qtbot.addWidget(window)
    window.show()
    return help_plugin


# =============================================================================
# Utility functions
# =============================================================================
def check_text(widget, text):
    """Check if some text is present in a widget."""
    if WEBENGINE:
        def callback(data):
            global html
            html = data
        widget.toHtml(callback)
        try:
            return text in html
        except NameError:
            return False
    else:
        return text in widget.toHtml()


def default_info_response():
    """Default response when asking for info."""
    return dict(name='', argspec='', note='', docstring='', calltip='',
                obj_text='')


# =============================================================================
# Tests
# =============================================================================
@flaky(max_runs=3)
def test_no_docs_message(help_plugin, qtbot):
    """
    Test that no docs message is shown when instrospection plugins
    can't get any info.
    """
    help_plugin.set_editor_doc(default_info_response())
    qtbot.waitUntil(lambda: check_text(help_plugin.get_widget()._webpage,
                                       "No documentation available"),
                    timeout=4000)


@flaky(max_runs=3)
def test_no_further_docs_message(help_plugin, qtbot):
    """
    Test that no further docs message is shown when instrospection
    plugins can get partial info.
    """
    info = default_info_response()
    info['name'] = 'foo'
    info['argspec'] = '(x, y)'

    help_plugin.set_editor_doc(info)
    qtbot.waitUntil(lambda: check_text(help_plugin.get_widget()._webpage,
                                       "No further documentation available"),
                    timeout=3000)


def test_help_opens_when_show_tutorial_unit(help_plugin, qtbot):
    """
    'Show tutorial' opens the help plugin if closed.

    Test fix for spyder-ide/spyder#6317.
    """
    help_plugin.switch_to_plugin = Mock()
    help_plugin.show_tutorial()
    assert help_plugin.switch_to_plugin.call_count == 1


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016 Pepijn Kenter.
# Copyright (c) 2019- Spyder Project Contributors
#
# Components of objectbrowser originally distributed under
# the MIT (Expat) license. Licensed under the terms of the MIT License;
# see NOTICE.txt in the Spyder root directory for details
# -----------------------------------------------------------------------------

# Standard library imports
from functools import lru_cache
import logging
from typing import Any, Callable, Optional

# Third-party imports
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import (
    QAbstractItemView, QActionGroup, QHeaderView, QTableWidget, QTreeView,
    QTreeWidget
)

# Local imports
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.config.base import _

logger = logging.getLogger(__name__)


# Toggle mixin
class ToggleColumnMixIn(SpyderWidgetMixin):
    """
    Adds actions to a QTableView that can show/hide columns
    by right clicking on the header
    """
    def add_header_context_menu(self, checked=None, checkable=None,
                                enabled=None):
        """
        Adds the context menu from using header information

        checked can be a header_name -> boolean dictionary. If given, headers
        with the key name will get the checked value from the dictionary.
        The corresponding column will be hidden if checked is False.

        checkable can be a header_name -> boolean dictionary. If given, headers
        with the key name will get the checkable value from the dictionary.

        enabled can be a header_name -> boolean dictionary. If given, headers
        with the key name will get the enabled value from the dictionary.
        """
        checked = checked if checked is not None else {}
        checkable = checkable if checkable is not None else {}
        enabled = enabled if enabled is not None else {}

        horizontal_header = self._horizontal_header()
        horizontal_header.setContextMenuPolicy(Qt.ActionsContextMenu)

        self.toggle_column_actions_group = QActionGroup(self)
        self.toggle_column_actions_group.setExclusive(False)
        self.__toggle_functions = []  # for keeping references

        for col in range(horizontal_header.count()):
            column_label = self.model().headerData(col, Qt.Horizontal,
                                                   Qt.DisplayRole)
            logger.debug("Adding: col {}: {}".format(col, column_label))
            func = self.__make_show_column_function(col)
            is_checked = checked.get(
                column_label,
                not horizontal_header.isSectionHidden(col)
            )
            action = self.create_action(
                name=f'show_{column_label}_column',
                text=str(column_label),
                tip=_("Shows or hides the {} column").format(column_label),
                toggled=func,
                initial=is_checked,
                parent=self.toggle_column_actions_group
            )
            action.setCheckable(checkable.get(column_label, True))
            action.setEnabled(enabled.get(column_label, True))
            self.__toggle_functions.append(func)  # keep reference
            horizontal_header.addAction(action)
            horizontal_header.setSectionHidden(col, not is_checked)

    def get_header_context_menu_actions(self):
        """Returns the actions of the context menu of the header."""
        return self._horizontal_header().actions()

    def _horizontal_header(self):
        """
        Returns the horizontal header (of type QHeaderView).

        Override this if the horizontalHeader() function does not exist.
        """
        return self.horizontalHeader()

    def __make_show_column_function(self, column_idx):
        """Creates a function that shows or hides a column."""
        show_column = lambda checked: self.setColumnHidden(column_idx,
                                                           not checked)
        return show_column

    def read_view_settings(self, key, settings=None, reset=False):
        """
        Reads the persistent program settings

        :param reset: If True, the program resets to its default settings
        :returns: True if the header state was restored, otherwise returns
                  False
        """
        logger.debug("Reading view settings for: {}".format(key))
        header_restored = False
        # TODO: Implement settings management
#        if not reset:
#            if settings is None:
#                settings = get_qsettings()
#            horizontal_header = self._horizontal_header()
#            header_data = settings.value(key)
#            if header_data:
#                header_restored = horizontal_header.restoreState(header_data)
#
#            # update actions
#            for col, action in enumerate(horizontal_header.actions()):
#                is_checked = not horizontal_header.isSectionHidden(col)
#                action.setChecked(is_checked)

        return header_restored

    def write_view_settings(self, key, settings=None):
        """Writes the view settings to the persistent store."""
        logger.debug("Writing view settings for: {}".format(key))
#       TODO: Settings management
#        if settings is None:
#            settings = get_qsettings()
#        settings.setValue(key, self._horizontal_header().saveState())


class ToggleColumnTableWidget(QTableWidget, ToggleColumnMixIn):
    """
    A QTableWidget where right clicking on the header allows the user
    to show/hide columns.
    """
    pass


class ToggleColumnTreeWidget(QTreeWidget, ToggleColumnMixIn):
    """
    A QTreeWidget where right clicking on the header allows the user to
    show/hide columns.
    """
    def _horizontal_header(self):
        """
        Returns the horizontal header (of type QHeaderView).

        Override this if the horizontalHeader() function does not exist.
        """
        return self.header()


class ToggleColumnTreeView(QTreeView, ToggleColumnMixIn):
    """
    A QTreeView where right clicking on the header allows the user to
    show/hide columns.
    """
    # Dummy conf section to avoid a warning
    CONF_SECTION = ""

    def __init__(
        self,
        namespacebrowser=None,
        data_function: Optional[Callable[[], Any]] = None,
        readonly=False
    ):
        QTreeView.__init__(self)
        self.readonly = readonly
        from spyder.plugins.variableexplorer.widgets.collectionsdelegate \
            import ToggleColumnDelegate
        self.delegate = ToggleColumnDelegate(
            self, namespacebrowser, data_function
        )
        self.setItemDelegate(self.delegate)
        self.setEditTriggers(QAbstractItemView.DoubleClicked)
        self.expanded.connect(self.resize_columns_to_contents)
        self.collapsed.connect(self.resize_columns_to_contents)

        # Dummy attribute to be compatible with BaseTableView
        self.hovered_row = -1

    @Slot()
    def resize_columns_to_contents(self):
        """Resize all the columns to its contents."""
        self._horizontal_header().resizeSections(QHeaderView.ResizeToContents)

    @lru_cache(maxsize=1)
    def selected_rows(self):
        """Dummy method to be compatible with BaseTableView."""
        return set()

    def _horizontal_header(self):
        """
        Returns the horizontal header (of type QHeaderView).

        Override this if the horizontalHeader() function does not exist.
        """
        return self.header()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#
"""Tests for autosaveerror.py"""

# Third party imports
from qtpy.QtWidgets import QPushButton

# Local imports
from spyder.plugins.editor.widgets.autosaveerror import AutosaveErrorDialog


def test_autosave_error_message_box(qtbot, mocker):
    """Test that AutosaveErrorDialog exec's at first, but that after the
    'do not show anymore' checkbox is clicked, it does not exec anymore."""
    mock_exec = mocker.patch.object(AutosaveErrorDialog, 'exec_')
    box = AutosaveErrorDialog('action', 'error')
    box.exec_if_enabled()
    assert mock_exec.call_count == 1
    box.dismiss_box.click()
    ok_button = box.findChild(QPushButton)
    ok_button.click()
    box2 = AutosaveErrorDialog('action', 'error')
    box2.exec_if_enabled()
    assert mock_exec.call_count == 1



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Namespace browser widget.

This is the main widget used in the Variable Explorer plugin
"""

# Standard library imports
import os
import os.path as osp
from pickle import UnpicklingError
import tarfile

# Third library imports
from qtpy import PYQT5, PYQT6
from qtpy.compat import getopenfilenames, getsavefilename
from qtpy.QtCore import Qt, Signal, Slot
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import (QApplication, QInputDialog, QMessageBox,
                            QVBoxLayout, QStackedLayout, QWidget)
from spyder_kernels.comms.commbase import CommError
from spyder_kernels.utils.iofuncs import iofunctions
from spyder_kernels.utils.misc import fix_reference_name
from spyder_kernels.utils.nsview import REMOTE_SETTINGS

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.config.utils import IMPORT_EXT
from spyder.widgets.collectionseditor import RemoteCollectionsEditorTableView
from spyder.plugins.variableexplorer.widgets.importwizard import ImportWizard
from spyder.utils import encoding
from spyder.utils.misc import getcwd_or_home, remove_backslashes
from spyder.widgets.helperwidgets import FinderWidget, PaneEmptyWidget


# Constants
VALID_VARIABLE_CHARS = r"[^\w+*=¡!¿?'\"#$%&()/<>\-\[\]{}^`´;,|¬]*\w"

# Max time before giving up when making a blocking call to the kernel
CALL_KERNEL_TIMEOUT = 30


class NamespaceBrowser(QWidget, SpyderWidgetMixin):
    """
    Namespace browser (global variables explorer widget).
    """
    # This is necessary to test the widget separately from its plugin
    CONF_SECTION = 'variable_explorer'

    # Signals
    sig_free_memory_requested = Signal()
    sig_start_spinner_requested = Signal()
    sig_stop_spinner_requested = Signal()
    sig_hide_finder_requested = Signal()

    sig_show_figure_requested = Signal(bytes, str, object)
    """
    This is emitted to request that a figure be shown in the Plots plugin.

    Parameters
    ----------
    image: bytes
        The image to show.
    mime_type: str
        The image's mime type.
    shellwidget: ShellWidget
        The shellwidget associated with the figure.
    """

    def __init__(self, parent):
        if PYQT5 or PYQT6:
            super().__init__(parent=parent, class_parent=parent)
        else:
            QWidget.__init__(self, parent)
            SpyderWidgetMixin.__init__(self, class_parent=parent)

        # Attributes
        self.filename = None
        self.plots_plugin_enabled = False

        # Widgets
        self.editor = None
        self.shellwidget = None
        self.finder = None
        self.pane_empty = PaneEmptyWidget(
            self,
            "variable-explorer",
            _("No variables to show"),
            _("Run code in the Editor or IPython console to see any "
              "global variables listed here for exploration and editing.")
        )

    def toggle_finder(self, show):
        """Show and hide the finder."""
        self.finder.set_visible(show)
        if not show:
            self.editor.setFocus()

    def do_find(self, text):
        """Search for text."""
        if self.editor is not None:
            self.editor.do_find(text)

    def finder_is_visible(self):
        """Check if the finder is visible."""
        if self.finder is None:
            return False
        return self.finder.isVisible()

    def setup(self):
        """
        Setup the namespace browser with provided options.
        """
        assert self.shellwidget is not None

        if self.editor is not None:
            self.set_namespace_view_settings()
            self.refresh_table()
        else:
            # Widgets
            self.editor = RemoteCollectionsEditorTableView(
                self,
                data=None,
                shellwidget=self.shellwidget,
                create_menu=False,
            )
            key_filter_dict = {
                Qt.Key_Up: self.editor.previous_row,
                Qt.Key_Down: self.editor.next_row
            }
            self.finder = FinderWidget(
                self,
                find_on_change=True,
                regex_base=VALID_VARIABLE_CHARS,
                key_filter_dict=key_filter_dict
            )

            # Signals
            self.editor.sig_files_dropped.connect(self.import_data)
            self.editor.sig_free_memory_requested.connect(
                self.sig_free_memory_requested)
            self.editor.sig_editor_creation_started.connect(
                self.sig_start_spinner_requested)
            self.editor.sig_editor_shown.connect(
                self.sig_stop_spinner_requested)

            self.finder.sig_find_text.connect(self.do_find)
            self.finder.sig_hide_finder_requested.connect(
                self.sig_hide_finder_requested)

            # Layout
            self.stack_layout = QStackedLayout()
            layout = QVBoxLayout()
            layout.setContentsMargins(0, 0, 0, 0)
            layout.setSpacing(0)
            layout.addWidget(self.editor)
            layout.addWidget(self.finder)

            self.table_widget = QWidget(self)
            self.table_widget.setLayout(layout)
            self.stack_layout.addWidget(self.table_widget)
            self.stack_layout.addWidget(self.pane_empty)
            self.setLayout(self.stack_layout)
            self.set_pane_empty()
            self.editor.source_model.sig_setting_data.connect(
                self.set_pane_empty)

    def set_pane_empty(self):
        if not self.editor.source_model.get_data():
            self.stack_layout.setCurrentWidget(self.pane_empty)
        else:
            self.stack_layout.setCurrentWidget(self.table_widget)

    def get_view_settings(self):
        """Return dict editor view settings"""
        settings = {}
        for name in REMOTE_SETTINGS:
            settings[name] = self.get_conf(name)

        return settings

    def set_shellwidget(self, shellwidget):
        """Bind shellwidget instance to namespace browser"""
        self.shellwidget = shellwidget

    def refresh_table(self):
        """Refresh variable table."""
        self.refresh_namespacebrowser()
        try:
            self.editor.resizeRowToContents()
            self.set_pane_empty()
        except TypeError:
            pass

    @Slot(dict)
    def update_view(self, kernel_state):
        """
        Update namespace view and other properties from a new kernel state.

        Parameters
        ----------
        kernel_state: dict
            A new kernel state. The structure of this dictionary is defined in
            the `SpyderKernel.get_state` method of Spyder-kernels.
        """
        if "namespace_view" in kernel_state:
            self.process_remote_view(kernel_state.pop("namespace_view"))
        if "var_properties" in kernel_state:
            self.set_var_properties(kernel_state.pop("var_properties"))

    def refresh_namespacebrowser(self, *, interrupt=True):
        """Refresh namespace browser"""
        if not self.shellwidget.spyder_kernel_ready:
            return
        self.shellwidget.call_kernel(
            interrupt=interrupt,
            callback=self.process_remote_view
        ).get_namespace_view()

        self.shellwidget.call_kernel(
            interrupt=interrupt,
            callback=self.set_var_properties
        ).get_var_properties()

    def set_namespace_view_settings(self):
        """Set the namespace view settings"""
        if not self.shellwidget.spyder_kernel_ready:
            return
        settings = self.get_view_settings()
        self.shellwidget.set_kernel_configuration(
            "namespace_view_settings", settings
        )

    def process_remote_view(self, remote_view):
        """Process remote view"""
        if remote_view is not None:
            self.set_data(remote_view)

    def set_var_properties(self, properties):
        """Set properties of variables"""
        if properties is not None:
            self.editor.var_properties = properties

    def set_data(self, data):
        """Set data."""
        if data != self.editor.source_model.get_data():
            self.editor.set_data(data)
            self.editor.adjust_columns()

    @Slot(list)
    def import_data(self, filenames=None):
        """Import data from text file."""
        title = _("Import data")
        if filenames is None:
            if self.filename is None:
                basedir = getcwd_or_home()
            else:
                basedir = osp.dirname(self.filename)
            filenames, _selfilter = getopenfilenames(self, title, basedir,
                                                     iofunctions.load_filters)
            if not filenames:
                return
        elif isinstance(filenames, str):
            filenames = [filenames]

        for filename in filenames:
            self.filename = str(filename)
            if os.name == "nt":
                self.filename = remove_backslashes(self.filename)
            extension = osp.splitext(self.filename)[1].lower()

            if extension == '.spydata':
                buttons = QMessageBox.Yes | QMessageBox.Cancel
                answer = QMessageBox.warning(
                    self,
                    title,
                    _("<b>Warning: %s files can contain malicious code!</b>"
                      "<br><br>"
                      "Do not continue unless this file is from a trusted "
                      "source. Would you like to import it "
                      "anyway?") % extension,
                    buttons
                )

                if answer == QMessageBox.Cancel:
                    return
            if extension not in iofunctions.load_funcs:
                buttons = QMessageBox.Yes | QMessageBox.Cancel
                answer = QMessageBox.question(
                    self,
                    title,
                    _("<b>Unsupported file extension '%s'</b>"
                      "<br><br>"
                      "Would you like to import it anyway by selecting a "
                      "known file format?") % extension,
                    buttons
                )
                if answer == QMessageBox.Cancel:
                    return
                formats = list(iofunctions.load_extensions.keys())
                item, ok = QInputDialog.getItem(self, title,
                                                _('Open file as:'),
                                                formats, 0, False)
                if ok:
                    extension = iofunctions.load_extensions[str(item)]
                else:
                    return

            load_func = iofunctions.load_funcs[extension]

            # 'import_wizard' (self.setup_io)
            if isinstance(load_func, str):
                # Import data with import wizard
                error_message = None
                try:
                    text, _encoding = encoding.read(self.filename)
                    base_name = osp.basename(self.filename)
                    editor = ImportWizard(self, text, title=base_name,
                                  varname=fix_reference_name(base_name))
                    if editor.exec_():
                        var_name, clip_data = editor.get_data()
                        self.editor.new_value(var_name, clip_data)
                except Exception as error:
                    error_message = str(error)
            else:
                QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
                QApplication.processEvents()
                error_message = self.load_data(self.filename, extension)
                QApplication.restoreOverrideCursor()
                QApplication.processEvents()

            if error_message is not None:
                QMessageBox.critical(self, title,
                                     _("<b>Unable to load '%s'</b>"
                                       "<br><br>"
                                       "The error message was:<br>%s"
                                       ) % (self.filename, error_message))
            self.refresh_table()

    def load_data(self, filename, ext):
        """Load data from a file."""
        if not self.shellwidget.spyder_kernel_ready:
            return
        overwrite = False
        if self.editor.var_properties:
            message = _('Do you want to overwrite old '
                        'variables (if any) in the namespace '
                        'when loading the data?')
            buttons = QMessageBox.Yes | QMessageBox.No
            result = QMessageBox.question(
                self, _('Data loading'), message, buttons)
            overwrite = result == QMessageBox.Yes
        try:
            return self.shellwidget.call_kernel(
                blocking=True,
                display_error=True,
                timeout=CALL_KERNEL_TIMEOUT).load_data(
                    filename, ext, overwrite=overwrite)
        except ImportError as msg:
            module = str(msg).split("'")[1]
            msg = _("Spyder is unable to open the file "
                    "you're trying to load because <tt>{module}</tt> is "
                    "not installed. Please install "
                    "this package in your working environment."
                    "<br>").format(module=module)
            return msg
        except TimeoutError:
            msg = _("Data is too big to be loaded")
            return msg
        except tarfile.ReadError:
            # Fixes spyder-ide/spyder#19126
            msg = _("The file could not be opened successfully. Recall that "
                    "the Variable Explorer supports the following file "
                    "extensions to import data:"
                    "<br><br><tt>{extensions}</tt>").format(
                        extensions=', '.join(IMPORT_EXT))
            return msg
        except TypeError:
            msg = _(
                "Spyder is unable to open the file you're trying to load. "
                "This could be caused due to a difference between the package "
                "versions used when you saved this spydata file and the ones "
                "installed in the current environment. Please check the "
                "compatibility between them (e.g. that you're using Numpy 2.x "
                "in both environments).<br>"
            )
        except (UnpicklingError, RuntimeError, CommError, OSError):
            return None

    def reset_namespace(self):
        warning = self.get_conf(
            section='ipython_console',
            option='show_reset_namespace_warning'
        )
        self.shellwidget.reset_namespace(warning=warning, message=True)
        self.editor.automatic_column_width = True

    def save_data(self):
        """Save data"""
        if not self.shellwidget.spyder_kernel_ready:
            return
        filename = self.filename
        if filename is None:
            filename = getcwd_or_home()
        extension = osp.splitext(filename)[1].lower()
        if not extension:
            # Needed to prevent trying to save a data file without extension
            # See spyder-ide/spyder#7196
            filename = filename + '.spydata'
        filename, _selfilter = getsavefilename(self, _("Save data"),
                                               filename,
                                               iofunctions.save_filters)
        if filename:
            self.filename = filename
        else:
            return False

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        QApplication.processEvents()

        error_message = self.save_namespace(self.filename)

        QApplication.restoreOverrideCursor()
        QApplication.processEvents()
        if error_message is not None:
            if 'Some objects could not be saved:' in error_message:
                save_data_message = (
                    _("<b>Some objects could not be saved:</b>")
                    + "<br><br><code>{obj_list}</code>".format(
                        obj_list=error_message.split(': ')[1]))
            else:
                save_data_message = _(
                    "<b>Unable to save current workspace</b>"
                    "<br><br>"
                    "The error message was:<br>") + error_message

            QMessageBox.critical(self, _("Save data"), save_data_message)

    def save_namespace(self, filename):
        try:
            return self.shellwidget.call_kernel(
                blocking=True,
                display_error=True,
                timeout=CALL_KERNEL_TIMEOUT).save_namespace(filename)
        except TimeoutError:
            msg = _("Data is too big to be saved")
            return msg
        except (UnpicklingError, RuntimeError, CommError):
            return None

    def plot(self, data, funcname):
        """
        Plot data.

        If all the following conditions are met:
        * the Plots plugin is enabled, and
        * the setting "Mute inline plotting" in the Plots plugin is set, and
        * the graphics backend in the IPython Console preferences is set
          to "inline",
        then call `plot_in_plots_plugin`, else call `plot_in_window`.
        """
        if (
            self.plots_plugin_enabled
            and self.get_conf('mute_inline_plotting', section='plots')
            and (
                self.get_conf('pylab/backend', section='ipython_console')
                == 'inline'
            )
        ):
            self.plot_in_plots_plugin(data, funcname)
        else:
            self.plot_in_window(data, funcname)

    def plot_in_plots_plugin(self, data, funcname):
        """
        Plot data in Plots plugin.

        Plot the given data to a PNG or SVG image and show the plot in the
        Plots plugin.
        """
        import spyder.pyplot as plt
        from IPython.core.pylabtools import print_figure

        try:
            from matplotlib import rc_context
        except ImportError:
            # Ignore fontsize and bottom options if guiqwt is used
            # as plotting library
            from contextlib import nullcontext as rc_context

        if self.get_conf('pylab/inline/figure_format',
                         section='ipython_console') == 1:
            figure_format = 'svg'
            mime_type = 'image/svg+xml'
        else:
            figure_format = 'png'
            mime_type = 'image/png'
        resolution = self.get_conf('pylab/inline/resolution',
                                   section='ipython_console')
        width = self.get_conf('pylab/inline/width',
                              section='ipython_console')
        height = self.get_conf('pylab/inline/height',
                               section='ipython_console')
        if self.get_conf('pylab/inline/bbox_inches',
                         section='ipython_console'):
            bbox_inches = 'tight'
        else:
            bbox_inches = None

        matplotlib_rc = {
            'font.size': self.get_conf('pylab/inline/fontsize',
                                       section='ipython_console'),
            'figure.subplot.bottom': self.get_conf('pylab/inline/bottom',
                                                   section='ipython_console')
        }

        with rc_context(matplotlib_rc):
            fig, ax = plt.subplots(figsize=(width, height))
            getattr(ax, funcname)(data)
            image = print_figure(
                fig,
                fmt=figure_format,
                bbox_inches=bbox_inches,
                dpi=resolution
            )

        if figure_format == 'svg':
            image = image.encode()
        self.sig_show_figure_requested.emit(image, mime_type, self.shellwidget)

    def plot_in_window(self, data, funcname):
        """
        Plot data in new Qt window.
        """
        import spyder.pyplot as plt

        plt.figure()
        getattr(plt, funcname)(data)
        plt.show()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2019- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the Spyder root directory for details)
# -----------------------------------------------------------------------------

"""
Tests for the array editor.
"""

# Standard library imports
from dataclasses import dataclass
import datetime
from unittest.mock import patch

# Third party imports
import numpy as np
import pytest
from qtpy.QtCore import Qt

# Local imports
from spyder.config.manager import CONF
from spyder.plugins.variableexplorer.widgets.objectexplorer import (
    ObjectExplorer)

# =============================================================================
# Fixtures
# =============================================================================
@pytest.fixture
def objectexplorer(qtbot):
    """Set up ObjectExplorer."""
    def create_objectexplorer(obj, **kwargs):
        editor = ObjectExplorer(obj, **kwargs)
        qtbot.addWidget(editor)
        return editor
    return create_objectexplorer


# =============================================================================
# Tests
# =============================================================================
def test_objectexplorer(objectexplorer):
    """Test to validate proper creation of the editor."""
    class Foobar(object):
        def __init__(self):
            self.text = "toto"

        def get_text(self):
            return self.text

        @property
        def error_attribute(self):
            # Attribute to test that the object explorer
            # handles errors by getattr gracefully
            raise AttributeError

    foobar = Foobar()

    # Editor was created
    editor = objectexplorer(foobar, name='foobar')
    assert editor

    # Check header data and default hidden sections
    header = editor.obj_tree.header()
    header_model = header.model()
    assert not header.isSectionHidden(0)
    assert header_model.headerData(0, Qt.Horizontal,
                                   Qt.DisplayRole) == "Name"
    assert not header.isSectionHidden(1)
    assert header_model.headerData(1, Qt.Horizontal,
                                   Qt.DisplayRole) == "Type"
    assert not header.isSectionHidden(2)
    assert header_model.headerData(2, Qt.Horizontal,
                                   Qt.DisplayRole) == "Size"
    assert not header.isSectionHidden(3)
    assert header_model.headerData(3, Qt.Horizontal,
                                   Qt.DisplayRole) == "Value"
    assert not header.isSectionHidden(4)
    assert header_model.headerData(4, Qt.Horizontal,
                                   Qt.DisplayRole) == "Callable"
    assert not header.isSectionHidden(5)
    assert header_model.headerData(5, Qt.Horizontal,
                                   Qt.DisplayRole) == "Path"
    assert header.isSectionHidden(6)
    assert header_model.headerData(6, Qt.Horizontal,
                                   Qt.DisplayRole) == "Id"
    assert header.isSectionHidden(7)
    assert header_model.headerData(7, Qt.Horizontal,
                                   Qt.DisplayRole) == "Attribute"
    assert header.isSectionHidden(8)
    assert header_model.headerData(8, Qt.Horizontal,
                                   Qt.DisplayRole) == "Routine"
    assert header.isSectionHidden(9)
    assert header_model.headerData(9, Qt.Horizontal,
                                   Qt.DisplayRole) == "File"
    assert header.isSectionHidden(10)
    assert header_model.headerData(10, Qt.Horizontal,
                                   Qt.DisplayRole) == "Source file"

    model = editor.obj_tree.model()

    # Check number of rows
    assert model.rowCount() == 1
    assert model.columnCount() == 11


@pytest.mark.parametrize('params', [
    # variable to show, rowCount for different Python 3 versions
    ('kjkj kj k j j kj k jkj', [71, 81]),
    ([1, 3, 4, 'kjkj', None], [45, 48]),
    ({1, 2, 1, 3, None, 'A', 'B', 'C', True, False}, [54, 57]),
    (1.2233, [57, 59]),
    (np.random.rand(10, 10), [162, 167]),
    (datetime.date(1945, 5, 8), [43, 48])
])
def test_objectexplorer_collection_types(objectexplorer, params):
    """Test to validate proper handling of collection data types."""
    test, row_count = params
    CONF.set('variable_explorer', 'show_special_attributes', True)

    # Editor was created
    editor = objectexplorer(test, name='variable')
    assert editor

    # Check number of rows and row content
    model = editor.obj_tree.model()

    # The row for the variable
    assert model.rowCount() == 1

    # Root row with children
    # Since rowCount for python 3 and 2 varies on differents systems,
    # we use a range of values
    expected_output_range = list(range(min(row_count), max(row_count) + 1))
    assert model.rowCount(model.index(0, 0)) in expected_output_range
    assert model.columnCount() == 11


@pytest.mark.parametrize('params', [
            # show_callable_, show_special_, rowCount for python 3 and 2
            (True, True, [35, 34, 26], ),  # 35: py3.11, 34: py3.x, 26: py2.7
            (False, False, [8, 8], )
        ])
def test_objectexplorer_types(objectexplorer, params):
    """Test to validate proper handling of data types inside an object."""
    class Foobar(object):
        def __init__(self):
            self.text = "toto"
            self.list = [1, 3, 4, 'kjkj', None]
            self.set = {1, 2, 1, 3, None, 'A', 'B', 'C', True, False},
            self.dict = {'d': 1, 'a': np.random.rand(10, 10), 'b': [1, 2]},
            self.float = 1.2233,
            self.array = np.random.rand(10, 10),
            self.date = datetime.date(1945, 5, 8),
            self.datetime = datetime.datetime(1945, 5, 8)
    foo = Foobar()

    show_callable, show_special, row_count = params
    CONF.set('variable_explorer', 'show_callable_attributes', show_callable)
    CONF.set('variable_explorer', 'show_special_attributes', show_special)

    # Editor was created
    editor = objectexplorer(foo, name='foo')
    assert editor

    # Check number of rows and row content
    model = editor.obj_tree.model()
    # The row for the object
    assert model.rowCount() == 1
    # Rows from the object attributes
    assert model.rowCount(model.index(0, 0)) in row_count
    assert model.columnCount() == 11


@dataclass
class DataclassForTesting:
    name: str
    price: float
    quantity: int


def test_objectexplorer_refreshaction_disabled():
    """
    Test that the refresh action is disabled by default.
    """
    data = DataclassForTesting('lemon', 0.15, 5)
    editor = ObjectExplorer(data, name='data')
    assert not editor.refresh_action.isEnabled()


def test_objectexplorer_refresh():
    """
    Test that after pressing the refresh button, the value of the editor is
    replaced by the return value of the data_function.
    """
    data_old = DataclassForTesting('lemon', 0.15, 5)
    data_new = range(1, 42, 3)
    editor = ObjectExplorer(data_old, name='data',
                            data_function=lambda: data_new)
    model = editor.obj_tree.model()
    root = model.index(0, 0)
    assert model.data(model.index(0, 0, root), Qt.DisplayRole) == 'name'
    assert model.data(model.index(0, 3, root), Qt.DisplayRole) == 'lemon'
    assert editor.refresh_action.isEnabled()
    editor.refresh_editor()
    model = editor.obj_tree.model()
    root = model.index(0, 0)
    row = model.rowCount(root) - 1
    assert model.data(model.index(row, 0, root), Qt.DisplayRole) == 'stop'
    assert model.data(model.index(row, 3, root), Qt.DisplayRole) == '42'


def test_objectexplorer_refresh_when_variable_deleted(qtbot):
    """
    Test that if the variable is deleted and then the editor is refreshed
    (resulting in data_function raising a KeyError), a critical dialog box
    is displayed and that the object editor is closed.
    """
    def datafunc():
        raise KeyError
    data = DataclassForTesting('lemon', 0.15, 5)
    editor = ObjectExplorer(data, name='data', data_function=datafunc)
    with patch('spyder.plugins.variableexplorer.widgets.objectexplorer'
               '.objectexplorer.QMessageBox.critical') as mock_critical:
        with qtbot.waitSignal(editor.rejected, timeout=0):
            editor.refresh_action.trigger()
    mock_critical.assert_called_once()


@dataclass
class Box:
    contents: object


def test_objectexplorer_refresh_nested():
    """
    Open an editor for a `Box` object containing a list, and then open another
    editor for the nested list. Test that refreshing the second editor works.
    """
    old_data = Box([1, 2, 3])
    new_data = Box([4, 5])
    editor = ObjectExplorer(
        old_data, name='data', data_function=lambda: new_data)
    model = editor.obj_tree.model()
    root_index = model.index(0, 0)
    contents_index = model.index(0, 0, root_index)
    editor.obj_tree.edit(contents_index)
    delegate = editor.obj_tree.delegate
    nested_editor = list(delegate._editors.values())[0]['editor']
    assert nested_editor.get_value() == [1, 2, 3]
    nested_editor.widget.refresh_action.trigger()
    assert nested_editor.get_value() == [4, 5]


def test_objectexplorer_refresh_doubly_nested():
    """
    Open an editor for a `Box` object containing another `Box` object which
    in turn contains a list. Then open a second editor for the nested list.
    Test that refreshing the second editor works.
    """
    old_data = Box(Box([1, 2, 3]))
    new_data = Box(Box([4, 5]))
    editor = ObjectExplorer(
        old_data, name='data', data_function=lambda: new_data)
    model = editor.obj_tree.model()
    root_index = model.index(0, 0)
    inner_box_index = model.index(0, 0, root_index)
    editor.obj_tree.expand(inner_box_index)
    contents_index = model.index(0, 0, inner_box_index)
    editor.obj_tree.edit(contents_index)
    delegate = editor.obj_tree.delegate
    nested_editor = list(delegate._editors.values())[0]['editor']
    assert nested_editor.get_value() == [1, 2, 3]
    nested_editor.widget.refresh_action.trigger()
    assert nested_editor.get_value() == [4, 5]


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

from spyder.plugins.ipythonconsole import SPYDER_KERNELS_VERSION
from spyder.config.base import running_remoteclient_tests
from spyder.plugins.remoteclient import SPYDER_REMOTE_VERSION


SERVER_ENV = "spyder-remote"
PACKAGE_NAME = "spyder-remote-services"
SCRIPT_URL = (
    f"https://raw.githubusercontent.com/spyder-ide/{PACKAGE_NAME}/main/scripts"
)


def get_installer_command(platform: str) -> str:
    if platform == "win":
        raise NotImplementedError("Windows is not supported yet")
    
    if running_remoteclient_tests():
        return '\n'  # server should be aready installed in the test environment

    return (
        f'"`which bash`" <(curl -L {SCRIPT_URL}/installer.sh) '
        f'"{SPYDER_REMOTE_VERSION}" "{SPYDER_KERNELS_VERSION}"'
    )


def get_server_version_command(platform: str) -> str:
    return (
        f"${{HOME}}/.local/bin/micromamba run -n {SERVER_ENV} python -c "
        "'import spyder_remote_services as sprs; print(sprs.__version__)'"
    )



# -*- coding: utf-8 -*-

# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Standard library imports
from unittest.mock import Mock, MagicMock

# Third party imports
from qtpy.QtWidgets import QMainWindow

# Local imports
from spyder.api.plugin_registration.registry import PLUGIN_REGISTRY
from spyder.config.manager import CONF
from spyder.plugins.completion.plugin import CompletionPlugin

# This is needed to avoid an error because QtAwesome
# needs a QApplication to work correctly.
from spyder.utils.qthelpers import qapplication
app = qapplication()

# PyTest imports
import pytest
from pytestqt.qtbot import QtBot


class MainWindowMock(QMainWindow):
    register_shortcut = Mock()

    def __init__(self):
        super().__init__(None)
        self.default_style = None
        self.widgetlist = []
        self.thirdparty_plugins = []
        self.shortcut_data = []
        self.prefs_dialog_instance = None
        self._APPLICATION_TOOLBARS = MagicMock()

        self.console = Mock()

        PLUGIN_REGISTRY.sig_plugin_ready.connect(self.register_plugin)

        # Load shortcuts for tests
        for context, name, __ in CONF.iter_shortcuts():
            self.shortcut_data.append((None, context, name, None, None))

        for attr in ['mem_status', 'cpu_status']:
            mock_attr = Mock()
            setattr(mock_attr, 'toolTip', lambda: '')
            setattr(mock_attr, 'setToolTip', lambda x: '')
            setattr(mock_attr, 'prefs_dialog_instance', lambda: '')
            setattr(self, attr, mock_attr)

    def register_plugin(self, plugin_name, external=False):
        plugin = PLUGIN_REGISTRY.get_plugin(plugin_name)
        plugin._register()

    def get_plugin(self, plugin_name, error=True):
        if plugin_name in PLUGIN_REGISTRY:
            return PLUGIN_REGISTRY.get_plugin(plugin_name)


@pytest.fixture(scope="module")
def qtbot_module(qapp, request):
    """Module fixture for qtbot."""
    result = QtBot(request)
    return result


def create_completion_plugin():
    @pytest.fixture(scope='module')
    def completion_plugin_wrap(qtbot_module, request):
        main_window = MainWindowMock()
        completions = CompletionPlugin(main_window, CONF)

        return completions
    return completion_plugin_wrap


completion_plugin_all = create_completion_plugin()


@pytest.fixture(scope='function')
def completion_plugin_all_started(request, qtbot_module,
                                  completion_plugin_all):

    completion_plugin = completion_plugin_all
    completion_plugin.wait_for_ms = 20000
    completion_plugin.start_all_providers()


    def wait_until_all_started():
        all_started = True
        for provider in completion_plugin.providers:
            
            provider_info = completion_plugin.providers[provider]
            all_started &= provider_info['status'] == completion_plugin.RUNNING
        return all_started

    qtbot_module.waitUntil(wait_until_all_started, timeout=30000)

    with qtbot_module.waitSignal(
            completion_plugin.sig_language_completions_available,
            timeout=30000) as blocker:
        completion_plugin.start_completion_services_for_language('python')

    capabilities, _ = blocker.args

    def teardown():
        completion_plugin.stop_all_providers()

    request.addfinalizer(teardown)
    return completion_plugin, capabilities



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# ----------------------------------------------------------------------------

"""Tests for status bar widget."""

# Standard library imports
import datetime

# Third-party imports
from flaky import flaky

# Local imports
from spyder.plugins.statusbar.widgets.tests.test_status import status_bar  # noqa
from spyder.plugins.application.widgets.status import InAppAppealStatus


@flaky(max_runs=5)
def test_inapp_appeal_status_bar(status_bar, qtbot):  # noqa
    """Test in-app appeal status bar widget."""
    plugin, window = status_bar
    widget = InAppAppealStatus(window)
    plugin.add_status_widget(widget)

    # The widget should be visible on a clean config
    assert widget.isVisible()

    # The last day the widget is shown should be today
    today = datetime.date.today()
    widget.get_conf("last_inapp_appeal") == str(today)

    # The widget should be visible during a day
    qtbot.wait(1000)
    widget.update_status()
    assert widget.isVisible()

    # The widget should be hidden if it was shown less than DAYS_TO_SHOW_AGAIN
    # ago
    some_days_ago = today - datetime.timedelta(
        days=widget.DAYS_TO_SHOW_AGAIN - 5
    )
    widget.set_conf("last_inapp_appeal", str(some_days_ago))
    widget.update_status()
    assert not widget.isVisible()

    # But it should be visible again if it was shown DAYS_TO_SHOW_AGAIN ago
    days_to_show_again = today - datetime.timedelta(
        days=widget.DAYS_TO_SHOW_AGAIN
    )
    widget.set_conf("last_inapp_appeal", str(days_to_show_again))
    widget.update_status()
    assert widget.isVisible()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Table widget to display a set of elements with title, description, icon and an
associated widget.
"""

# Standard library imports
import sys
from typing import List, Optional, TypedDict

# Third-party imports
import qstylizer.style
from qtpy.QtCore import QAbstractTableModel, QModelIndex, QSize, Qt
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QAbstractItemView, QCheckBox, QHBoxLayout, QWidget
from superqt.utils import qdebounced

# Local imports
from spyder.api.fonts import SpyderFontsMixin, SpyderFontType
from spyder.utils.icon_manager import ima
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle
from spyder.widgets.helperwidgets import HoverRowsTableView, HTMLDelegate


class Element(TypedDict):
    """Spec for elements that can be displayed in ElementsTable."""

    title: str
    """Element title"""

    description: str
    """Element description"""

    additional_info: Optional[str]
    """
    Additional info that needs to be displayed in a separate column (optional)
    """

    icon: Optional[QIcon]
    """Element icon (optional)"""

    widget: Optional[QWidget]
    """
    Element widget, e.g. a checkbox or radio button associated to the element
    (optional)
    """


class ElementsModel(QAbstractTableModel, SpyderFontsMixin):

    def __init__(
        self,
        parent: QWidget,
        elements: List[Element],
        with_icons: bool,
        with_addtional_info: bool,
        with_widgets: bool,
    ):
        QAbstractTableModel.__init__(self)

        self.elements = elements
        self.with_icons = with_icons

        # Number of columns
        self.n_columns = 1

        # Index corresponding to columns. The 'title' column is always expected
        self.columns = {'title': 0}

        # Extra columns
        if with_addtional_info:
            self.n_columns += 1
            self.columns['additional_info'] = 1

        if with_widgets:
            self.n_columns += 1

            if self.n_columns == 3:
                self.columns['widgets'] = 2
            else:
                self.columns['widgets'] = 1

        # Text styles
        text_color = SpyderPalette.COLOR_TEXT_1
        title_font_size = self.get_font(
            SpyderFontType.Interface, font_size_delta=1).pointSize()

        self.title_style = f'color:{text_color}; font-size:{title_font_size}pt'
        self.additional_info_style = f'color:{SpyderPalette.COLOR_TEXT_4}'
        self.description_style = f'color:{text_color}'

    # ---- Qt overrides
    # -------------------------------------------------------------------------
    def data(self, index, role=Qt.DisplayRole):

        element = self.elements[index.row()]

        if role == Qt.DisplayRole:
            if index.column() == self.columns['title']:
                return self.get_title_repr(element)
            elif index.column() == self.columns.get('additional_info'):
                return self.get_info_repr(element)
            else:
                return None
        elif role == Qt.DecorationRole and self.with_icons:
            if index.column() == self.columns['title']:
                return element['icon']
            else:
                return None

        return None

    def rowCount(self, index=QModelIndex()):
        return len(self.elements)

    def columnCount(self, index=QModelIndex()):
        return self.n_columns

    # ---- Own methods
    # -------------------------------------------------------------------------
    def get_title_repr(self, element: Element) -> str:
        return (
            f'<table cellspacing="0" cellpadding="3">'
            # Title
            f'<tr><td><span style="{self.title_style}">'
            f'{element["title"]}'
            f'</span></td></tr>'
            # Description
            f'<tr><td><span style="{self.description_style}">'
            f'{element["description"]}'
            f'</span></td></tr>'
            f'</table>'
        )

    def get_info_repr(self, element: Element) -> str:
        if element.get('additional_info'):
            additional_info = f' {element["additional_info"]}'
        else:
            return ''

        return (
            f'<span style="{self.additional_info_style}">'
            f'{additional_info}'
            f'</span>'
        )


class ElementsTable(HoverRowsTableView):

    def __init__(self, parent: Optional[QWidget], elements: List[Element]):
        HoverRowsTableView.__init__(self, parent, custom_delegate=True)
        self.elements = elements

        # Check for additional features
        self._with_icons = self._with_feature('icon')
        self._with_addtional_info = self._with_feature('additional_info')
        self._with_widgets = self._with_feature('widget')

        # To keep track of the current row widget (e.g. a checkbox) in order to
        # change its background color when its row is hovered.
        self._current_row = -1
        self._current_row_widget = None

        # To make adjustments when the widget is shown
        self._is_shown = False

        # This is used to paint the entire row's background color when its
        # hovered.
        self.sig_hover_index_changed.connect(self._on_hover_index_changed)

        # Set model
        self.model = ElementsModel(
            self,
            self.elements,
            self._with_icons,
            self._with_addtional_info,
            self._with_widgets
        )
        self.setModel(self.model)

        # Adjustments for the title column
        title_delegate = HTMLDelegate(self, margin=9, wrap_text=True)
        self.setItemDelegateForColumn(
            self.model.columns['title'], title_delegate)
        self.sig_hover_index_changed.connect(
             title_delegate.on_hover_index_changed)

        # Adjustments for the additional info column
        self._info_column_width = 0
        if self._with_addtional_info:
            info_delegate = HTMLDelegate(self, margin=10, align_vcenter=True)
            self.setItemDelegateForColumn(
                self.model.columns['additional_info'], info_delegate)
            self.sig_hover_index_changed.connect(
                 info_delegate.on_hover_index_changed)

            # This is necessary to get this column's width below
            self.resizeColumnsToContents()

            self._info_column_width = self.horizontalHeader().sectionSize(
                self.model.columns['additional_info'])

        # Adjustments for the widgets column
        self._widgets_column_width = 0
        if self._with_widgets:
            widgets_delegate = HTMLDelegate(self, margin=0)
            self.setItemDelegateForColumn(
                self.model.columns['widgets'], widgets_delegate)
            self.sig_hover_index_changed.connect(
                 widgets_delegate.on_hover_index_changed)

            # This is necessary to get this column's width below
            self.resizeColumnsToContents()

            # Note: We add 15 pixels to the Qt width so that the widgets are
            # not so close to the right border of the table, which doesn't look
            # good.
            self._widgets_column_width = self.horizontalHeader().sectionSize(
                self.model.columns['widgets']) + 15

            # Add widgets
            for i in range(len(self.elements)):
                layout = QHBoxLayout()
                layout.addWidget(self.elements[i]['widget'])
                layout.setAlignment(Qt.AlignHCenter)

                container_widget = QWidget(self)
                container_widget.setLayout(layout)

                # This key is not accounted for in Element because it's only
                # used internally, so it doesn't need to provided in a list of
                # Element's.
                self.elements[i]['row_widget'] = container_widget

                self.setIndexWidget(
                    self.model.index(i, self.model.columns['widgets']),
                    container_widget
                )

        # Make last column take the available space to the right
        self.horizontalHeader().setStretchLastSection(True)

        # Hide headers
        self.horizontalHeader().hide()
        self.verticalHeader().hide()

        # Set icons size
        if self._with_icons:
            self.setIconSize(QSize(32, 32))

        # Hide grid to only paint horizontal lines with css
        self.setShowGrid(False)

        # Set selection behavior
        self.setSelectionMode(QAbstractItemView.NoSelection)

        # Set stylesheet
        self._set_stylesheet()

    # ---- Private API
    # -------------------------------------------------------------------------
    def _on_hover_index_changed(self, index):
        """Actions to take when the index that is hovered has changed."""
        row = index.row()

        if row != self._current_row:
            self._current_row = row

            if self._with_widgets:
                # Remove background color of previous row widget
                if self._current_row_widget is not None:
                    self._current_row_widget.setStyleSheet("")

                # Set background for the new row widget
                new_row_widget = self.elements[row]["row_widget"]
                new_row_widget.setStyleSheet(
                    f"background-color: {SpyderPalette.COLOR_BACKGROUND_3}"
                )

                # Set new current row widget
                self._current_row_widget = new_row_widget

    def _set_stylesheet(self, leave=False):
        """Set stylesheet when entering or leaving the widget."""
        css = qstylizer.style.StyleSheet()
        bgcolor = SpyderPalette.COLOR_BACKGROUND_1 if leave else "transparent"

        css["QTableView::item"].setValues(
            borderBottom=f"1px solid {SpyderPalette.COLOR_BACKGROUND_4}",
            paddingLeft="5px",
            backgroundColor=bgcolor
        )

        self.setStyleSheet(css.toString())

    def _set_layout(self):
        """
        Set rows and columns layout.

        This is necessary to make the table look good at different sizes.
        """
        # We need to make these extra adjustments for Mac so that the last
        # column is not too close to the right border
        extra_width = 0
        if sys.platform == 'darwin':
            if self.verticalScrollBar().isVisible():
                extra_width = (
                    AppStyle.MacScrollBarWidth +
                    (15 if self._with_widgets else 5)
                )
            else:
                extra_width = 10 if self._with_widgets else 5

        # Resize title column so that the table fits into the available
        # horizontal space.
        if self._info_column_width > 0 or self._widgets_column_width > 0:
            title_column_width = (
                self.horizontalHeader().size().width() -
                (self._info_column_width + self._widgets_column_width +
                 extra_width)
            )

            self.horizontalHeader().resizeSection(
                self.model.columns['title'], title_column_width
            )

        # Resize rows. This is done because wrapping text in HTMLDelegate's
        # changes row heights in unpredictable ways.
        self.resizeRowsToContents()

    _set_layout_debounced = qdebounced(_set_layout, timeout=40)
    """
    Debounced version of _set_layout.

    Notes
    -----
    * We need a different version of _set_layout so that we can use the regular
      one in showEvent. That way users won't experience a visual glitch when
      the widget is rendered for the first time.
    * We use this version in resizeEvent, where that is not a problem.
    """

    def _with_feature(self, feature_name: str) -> bool:
        """Check if it's necessary to build the table with `feature_name`."""
        return len([e for e in self.elements if e.get(feature_name)]) > 0

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def showEvent(self, event):
        if not self._is_shown:
            self._set_layout()

            # To not run the adjustments above every time the widget is shown
            self._is_shown = True

        super().showEvent(event)

    def leaveEvent(self, event):
        super().leaveEvent(event)

        # Clear background color painted on hovered row widget
        if self._current_row_widget is not None:
            self._current_row_widget.setStyleSheet('')
        self._set_stylesheet(leave=True)

    def enterEvent(self, event):
        super().enterEvent(event)

        # Restore background color that's going to be painted on hovered row
        if self._current_row_widget is not None:
            self._current_row_widget.setStyleSheet(
                f"background-color: {SpyderPalette.COLOR_BACKGROUND_3}"
            )
        self._set_stylesheet()

    def resizeEvent(self, event):
        # This is necessary to readjust the layout when the parent widget is
        # resized.
        self._set_layout_debounced()
        super().resizeEvent(event)


def test_elements_table():
    from spyder.utils.qthelpers import qapplication
    app = qapplication()  # noqa

    elements_with_title = [
        {'title': 'IPython console', 'description': 'Execute code'},
        {'title': 'Help', 'description': 'Look for help'}
    ]

    table = ElementsTable(None, elements_with_title)
    table.show()

    elements_with_icons = [
        {'title': 'IPython console', 'description': 'Execute code',
         'icon': ima.icon('ipython_console')},
        {'title': 'Help', 'description': 'Look for help',
         'icon': ima.icon('help')}
    ]

    table_with_icons = ElementsTable(None, elements_with_icons)
    table_with_icons.show()

    elements_with_widgets = [
        {'title': 'IPython console', 'description': 'Execute code',
         'icon': ima.icon('ipython_console'), 'widget': QCheckBox()},
        {'title': 'Help', 'description': 'Look for help',
         'icon': ima.icon('help'), 'widget': QCheckBox()}
    ]

    table_with_widgets = ElementsTable(None, elements_with_widgets)
    table_with_widgets.show()

    elements_with_info = [
        {'title': 'IPython console', 'description': 'Execute code',
         'icon': ima.icon('ipython_console'), 'widget': QCheckBox(),
         'additional_info': 'Core plugin'},
        {'title': 'Help', 'description': 'Look for help',
         'icon': ima.icon('help'), 'widget': QCheckBox()}
    ]

    table_with_widgets_and_icons = ElementsTable(None, elements_with_info)
    table_with_widgets_and_icons.show()

    app.exec_()


if __name__ == '__main__':
    test_elements_table()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Script and functions to automatically remove trailing spaces from files."""

# Standard library imports
import argparse
import datetime
import re
import subprocess
import sys
import warnings

# Third party imports
import requests


PR_DATE_CUTOFF = datetime.datetime(2018, 12, 8, 0)
INCLUDE_EXTENSIONS = {"bat", "c", "cpp", "css", "csv", "gitignore",
                      "gitattributes", "html", "in", "ini", "js", "md", "py",
                      "pyx", "qss", "R", "rst", "sh", "txt", "xml", "yml"}
EXCLUDE_PATHS = [
    "/tests?/(?!(tests?_|__init__|fixture))",  # Skip all test datafiles
    "/mathjax/",  # Mathjax is vendored code and should be left as is
    r"^\.github/ISSUE_TEMPLATE\.md$",  # Ws needed before bullets
    r"^spyder/defaults/.*\.ini$",  # Could break ancient Spyder versions
    "^CHANGELOG.md$",  # Don't remove ws in the changelog, for some reason
    ]


def git_files(file_types):
    """Get index, staged, or modified files from git."""
    files_final = set()
    git_commands = {
        "staged": ("git", "diff", "--name-only", "--cached",
                   "--diff-filter=d", "-z"),
        "unstaged": ("git", "diff", "--name-only", "--diff-filter=d", "-z"),
        "untracked": ("git", "ls-files", "-o", "--exclude-standard", "-z"),
        "index": ("git", "ls-files", "--cached", "--exclude-standard", "-z"),
        }

    for file_type in file_types:
        try:
            status_output = subprocess.run(git_commands[file_type],
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE,
                                           check=True,
                                           encoding="utf-8")
        # If not inside the repo or git otherwise runs into a problem
        except subprocess.CalledProcessError as e:
            raise RuntimeError("Must be in the root of a Git repo") from e
        # If subprocess.run doesn't exist or doesn't have the encoding arg
        except (TypeError, AttributeError):
            raise RuntimeError("Python >=3.6 required to run script.")
        if status_output.stdout:
            files_final = files_final.union(
                set(status_output.stdout.strip("\0").split("\0")))
        else:
            files_final = set()
    return files_final


def filter_files(file_paths, include_extensions=INCLUDE_EXTENSIONS,
                 exclude_paths=EXCLUDE_PATHS):
    """Filter files by extension and path."""
    filtered_paths = {file_path for file_path in file_paths
                      if file_path.split(".")[-1] in include_extensions}
    exclude_paths_regex = re.compile("|".join(exclude_paths))
    filtered_paths = {file_path for file_path in filtered_paths
                      if not exclude_paths_regex.search(file_path)}
    return filtered_paths


def get_github_prs(repo_path=None):
    """Get the files touched by open PRs on the given repo."""
    pr_api_url = "https://api.github.com/repos/{repo_path}/pulls"

    if repo_path is None:
        status_output = subprocess.run(
            ["git", "remote", "get-url", "upstream"],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")
        if not status_output.stdout:
            status_output = subprocess.run(
                ["git", "remote", "get-url", "origin"],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                encoding="utf-8")
            if not status_output.stdout:
                raise ValueError("Repo path not provided and could not be "
                                 "determined via git remote origin/upstream.")
        repo_path = "/".join(
            status_output.stdout.split(":")[1].split("/")[-2:]).split(".")[0]

    pr_api_url = pr_api_url.format(repo_path=repo_path)
    pr_params = {"status": "open", "per_page": 99}
    pr_responce = requests.get(pr_api_url, pr_params)
    pr_responce.raise_for_status()
    open_prs = pr_responce.json()

    open_prs = [pr for pr in open_prs if datetime.datetime.strptime(
        pr["created_at"], "%Y-%m-%dT%H:%M:%SZ") >= PR_DATE_CUTOFF]

    return open_prs


def checkout_branch(branch="master"):
    """Checkout the given branch, creating it if it does not already exist."""
    status_output = subprocess.run(
        ["git", "rev-parse", "--abbrev-ref", "HEAD"],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")
    previous_branch = status_output.stdout.strip()
    status_output = subprocess.run(["git", "rev-parse", "--verify", branch],
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   encoding="utf-8")
    if status_output.stdout:
        subprocess.run(["git", "checkout", branch])
    else:
        subprocess.run(["git", "checkout", "-b", branch])

    return previous_branch


def get_pr_merge_conflicts(prs_tocheck, branch_to_compare="master"):
    """Determine the files that would have conflicts with a certain branch."""
    merge_conflicts_bypr = {}
    previous_branch = checkout_branch(branch=branch_to_compare)

    for pr in prs_tocheck:
        try:
            subprocess.run(["git", "pr", str(pr["number"])], check=True)
        except subprocess.CalledProcessError:
            warnings.warn("git pr failed; installing git pr alias.")
            subprocess.run(
                ["git", "config", "--local", "alias.pr",
                 ("!f() { git fetch -fu "
                  "${2:-$(git remote |grep ^upstream || echo origin)} "
                  "refs/pull/$1/head:pr/$1 && git checkout pr/$1; }; f")])
            subprocess.run(["git", "pr", str(pr["number"])], check=True)

        subprocess.run(["git", "checkout", branch_to_compare])
        subprocess.run(["git", "merge", "--no-commit", "--no-ff",
                        "pr/" + str(pr["number"])])
        status_output = subprocess.run(
            ["git", "ls-files", "-u", "--exclude-standard", "-z"],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")
        if status_output.stdout:
            merge_conflicts = {file_data.split("\t")[-1] for file_data in
                               status_output.stdout.strip("\0").split("\0")}
        else:
            merge_conflicts = set()
        subprocess.run(["git", "merge", "--abort"])
        merge_conflicts_bypr[pr["number"]] = merge_conflicts

    try:
        subprocess.run(["git", "pr-clean"], check=True)
    except subprocess.CalledProcessError:
        warnings.warn("git pr-clean failed; installing git pr-clean alias.")
        subprocess.run(
            ["git", "config", "--local", "alias.pr-clean",
             ("!git for-each-ref refs/heads/pr/* --format='%(refname)' "
              "| while read ref ; do branch=${ref#refs/heads/} ; "
              "git branch -D $branch ; done")])
        subprocess.run(["git", "pr-clean"])
    subprocess.run(["git", "checkout", previous_branch])
    return merge_conflicts_bypr


def handle_whitespace(file_paths, whitespace_type="all", warn_only=False,
                      verbose=False):
    """Detect or eliminate whitespace in the passed files."""
    if verbose is False and warn_only:
        verbose = 2
    if verbose:
        print("{action} {ws_type} whitespace in the following files:".format(
            action=("Detected" if warn_only else "Removed"),
            ws_type=whitespace_type))
    found_files = {}
    if whitespace_type == "blank":
        space_regex = (re.compile(r"\n[ \t]+\n"), r"\n\n")
    elif whitespace_type == "trailing":
        space_regex = (re.compile(r"([^ \t])[ \t]+\n"), r"\1\n")
    elif whitespace_type == "both":
        space_regex = (re.compile(r"[ \t]+\n"), r"\n")
    else:
        raise ValueError(
            "handle_type must be one of 'blank', 'trailing' or 'all'")

    for file_path in file_paths:
        try:
            with open(file_path, mode="r", encoding="utf-8") as file:
                file_contents = file.read()
                newline_char = file.newlines
        except (OSError, UnicodeError, IOError) as e:
            warnings.warn("{e_type} reading file {fname}: {e_msg}".format(
                e_type=type(e), fname=file_path, e_msg=str(e)))
            continue
        output_contents, n_subs = re.subn(*space_regex, file_contents)
        if n_subs:
            if verbose >= 2:
                print("{ws_count} : {file_path}".format(
                    ws_count=n_subs, file_path=file_path))
            if not warn_only:
                try:
                    with open(file_path, mode="w", encoding="utf-8",
                              newline=newline_char) as file:
                        file_contents = file.write(output_contents)
                except (OSError, UnicodeError, IOError) as e:
                    warnings.warn("{e_type} reading file {fname}: {e_msg}"
                                  .format(e_type=type(e), fname=file_path,
                                          e_msg=str(e)))
                    continue
            found_files[file_path] = n_subs
    if verbose:
        if found_files:
            print("{n_ws} instances in {n_files} file(s)".format(
                n_ws=sum(found_files.values()),
                n_files=len(found_files)))
        else:
            print("No matching whitespace found.")

    return found_files


def handle_whitespace_files(file_types, whitespace_type="all",
                            warn_only=False, verbose=False,
                            check_prs=False):
    """Remove trailing whitespace in all or selected files in the project."""
    files_toprocess = git_files(file_types)
    files_toprocess = filter_files(files_toprocess)

    if check_prs:
        open_prs = get_github_prs()
        conflicts_before = get_pr_merge_conflicts(open_prs, "master")
        temp_branch_name = "temp-test-nospace"
        previous_branch = checkout_branch(branch=temp_branch_name)
        handle_whitespace(files_toprocess, whitespace_type=whitespace_type,
                          warn_only=False, verbose=0)
        subprocess.run(["git", "commit", "-a", "-m", "test"])
        conflicts_after = get_pr_merge_conflicts(open_prs, temp_branch_name)

        all_conflicts = set()
        conflicts_bypr = {}
        for pr in open_prs:
            whitespace_conflicts = conflicts_after[pr["number"]].difference(
                conflicts_before[pr["number"]])
            if whitespace_conflicts:
                conflicts_bypr[pr["number"]] = whitespace_conflicts
                all_conflicts = all_conflicts.union(whitespace_conflicts)
                if verbose >= 2:
                    print("PR #{pr_num} has conflicts {conf}".format(
                        pr_num=pr["number"], conf=whitespace_conflicts))
        if verbose:
            print("Total conflicts: {num_conf} in {num_prs} PRs".format(
                num_conf=len(all_conflicts), num_prs=len(conflicts_bypr)))
        subprocess.run(["git", "checkout", previous_branch])
        subprocess.run(["git", "branch", "-D", temp_branch_name])
        files_toprocess = files_toprocess.difference(all_conflicts)

    found_files = handle_whitespace(
        files_toprocess, whitespace_type=whitespace_type,
        warn_only=warn_only, verbose=verbose)

    if check_prs:
        found_files = (found_files, conflicts_bypr)

    return found_files


def generate_arg_parser():
    """Generate the argument parser for the trailing whitespace script."""
    description = "Automatically remove trailing whitespace from files."
    arg_parser = argparse.ArgumentParser(description=description)

    arg_parser.add_argument(
        "--whitespace-type", default="both",
        help=("Type of whitespace to remove. 'blank' for just whitespace on "
              "otherwise-blank lines, 'trailing' for whitespace on non-empty "
              "lines, and 'both' to remove both (the default)."))

    arg_parser.add_argument(
        "--file-types", default=["staged", "unstaged", "untracked"], nargs="*",
        help=("Type(s) of files to process, as determined by git. "
              "Options are 'staged', 'unstaged', 'untracked' and 'index' "
              "(to process the entire git index). "
              "By default, does the first 3 to capture all modified files, "
              "whether or not they are staged."))

    arg_parser.add_argument(
        "--warn-only", action="store_true",
        help=("If passed, will only warn (and exit with a non-zero status) "
              "if trailing spaces are found in the files, rather than "
              "removing them. Intended for a CI-side check. "
              "Also implictly triggers verbosity=2."))

    arg_parser.add_argument(
        "--verbose", "-v", action="count",
        help=("How much detail to print about the files converted. "
              "At verbosity level 0 (the default), nothing is printed. "
              "At level 1 (-v), prints a summary of trailing spaces and "
              "conflicts detected across all files. At level 2, (-vv), "
              "prints detailed information about every file and conflict."))

    arg_parser.add_argument(
        "--check-prs", action="store_true",
        help=("If passed, will check the changes to see if they cause "
              "merge conflicts on any open PRs on the repo, and only "
              "make them in the files that do not result in such. "
              "Installs the ``pr`` and ``pr-clean`` git alises if not found."))

    return arg_parser


if __name__ == "__main__":
    arg_parser = generate_arg_parser()
    script_args = arg_parser.parse_args()
    found_files = handle_whitespace_files(**script_args.__dict__)
    if found_files and script_args.warn_only:
        sys.exit(1)
    else:
        sys.exit(0)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for the indentation feature
"""

# Third party imports
import pytest
from qtpy.QtGui import QTextCursor

# Local imports
from spyder.py3compat import to_text_string
from spyder.plugins.editor.widgets.codeeditor import CodeEditor


# --- Helper methods
# -----------------------------------------------------------------------------
def make_indent(editor, single_line=True, start_line=1):
    """Indent and return code."""
    editor.go_to_line(start_line)
    if not single_line:
        editor.moveCursor(QTextCursor.End, mode=QTextCursor.KeepAnchor)
    editor.indent()
    text = editor.toPlainText()
    return to_text_string(text)


def make_unindent(editor, single_line=True, start_line=1):
    """Unindent and return code."""
    editor.go_to_line(start_line)
    if not single_line:
        editor.moveCursor(QTextCursor.End, mode=QTextCursor.KeepAnchor)
    editor.unindent()
    text = editor.toPlainText()
    return to_text_string(text)

# --- Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture
def codeeditor_indent(codeeditor):
    """
    Setup CodeEditor with some text useful for folding related tests.
    """
    editor = codeeditor
    editor.set_indent_chars(" " * 2)
    return editor


# --- Tests
# -----------------------------------------------------------------------------
def test_single_line_indent(codeeditor_indent):
    """Test indentation in a single line."""
    editor = codeeditor_indent
    text = ("class a():\n"
            "self.b = 1\n"
            "print(self.b)\n"
            "\n"
            )
    expected = ("class a():\n"
                "  self.b = 1\n"
                "print(self.b)\n"
                "\n"
                )
    editor.set_text(text)
    # Indent line without spaces
    new_text = make_indent(editor, start_line=2)
    assert new_text == expected


def test_selection_indent(codeeditor_indent):
    """Test indentation with selection of more than one line."""
    editor = codeeditor_indent
    text = ("class a():\n"
            "self.b = 1\n"
            "print(self.b)\n"
            "\n"
            )
    expected = ("class a():\n"
                "  self.b = 1\n"
                "  print(self.b)\n"
                "  \n"
                )

    editor.set_text(text)
    # Toggle manually commented code
    new_text = make_indent(editor, single_line=False, start_line=2)
    assert new_text == expected


def test_fix_indentation(codeeditor_indent):
    """Test fix_indentation() method."""
    editor = codeeditor_indent
    # Contains tabs.
    original = ("\t\n"
                "class a():\t\n"
                "\tself.b = 1\n"
                "\tprint(self.b)\n"
                "\n"
                )
    # Fix indentation replaces tabs with indent_chars spaces.
    fixed = ("  \n"
             "class a():  \n"
             "  self.b = 1\n"
             "  print(self.b)\n"
             "\n"
             )
    editor.set_text(original)
    editor.fix_indentation()
    assert to_text_string(editor.toPlainText()) == fixed
    assert editor.document().isModified()
    # Test that undo/redo works - spyder-ide/spyder#1754.
    editor.undo()
    assert to_text_string(editor.toPlainText()) == original
    assert not editor.document().isModified()
    editor.redo()
    assert to_text_string(editor.toPlainText()) == fixed
    assert editor.document().isModified()


def test_single_line_unindent(codeeditor_indent):
    """Test unindentation in a single line."""
    editor = codeeditor_indent
    text = ("class a():\n"
            "  self.b = 1\n"
            "print(self.b)\n"
            "\n"
            )
    expected = ("class a():\n"
                "self.b = 1\n"
                "print(self.b)\n"
                "\n"
                )
    editor.set_text(text)
    # Indent line without spaces
    new_text = make_unindent(editor, start_line=2)
    assert new_text == expected


def test_selection_unindent(codeeditor_indent):
    """Test unindentation with selection of more than one line."""
    editor = codeeditor_indent
    text = ("class a():\n"
            "  self.b = 1\n"
            "  print(self.b)\n"
            "  \n"
            )
    expected = ("class a():\n"
                "self.b = 1\n"
                "print(self.b)\n"
                "\n"
                )
    editor.set_text(text)
    # Toggle manually commented code
    new_text = make_unindent(editor, single_line=False, start_line=2)
    assert new_text == expected


def test_single_line_unindent_to_grid(codeeditor_indent):
    """Test unindentation in a single line."""
    editor = codeeditor_indent
    text = ("class a():\n"
            "   self.b = 1\n"
            "print(self.b)\n"
            "\n"
            )
    expected = ("class a():\n"
                "  self.b = 1\n"
                "print(self.b)\n"
                "\n"
                )
    editor.set_text(text)
    # Indent line without spaces
    new_text = make_unindent(editor, start_line=2)
    assert new_text == expected

    expected2 = ("class a():\n"
                 "self.b = 1\n"
                 "print(self.b)\n"
                 "\n"
                 )
    new_text2 = make_unindent(editor, start_line=2)
    assert new_text2 == expected2


def test_selection_unindent_to_grid(codeeditor_indent):
    """Test unindentation with selection of more than one line."""
    editor = codeeditor_indent
    text = ("class a():\n"
            "   self.b = 1\n"
            "   print(self.b)\n"
            "\n"
            )
    expected = ("class a():\n"
                "  self.b = 1\n"
                "  print(self.b)\n"
                "\n"
                )
    editor.set_text(text)
    # Toggle manually commented code
    new_text = make_unindent(editor, single_line=False, start_line=2)
    assert new_text == expected

    expected2 = ("class a():\n"
                 "self.b = 1\n"
                 "print(self.b)\n"
                 "\n"
                 )
    new_text2 = make_unindent(editor, single_line=False, start_line=2)
    assert new_text2 == expected2


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for conda.py"""

# Standard library imports
import os
import sys
import time

# Third party imports
import pytest

# Local imports
from spyder.config.base import running_in_ci
from spyder.config.utils import is_anaconda
from spyder.plugins.ipythonconsole.tests.conftest import get_conda_test_env
from spyder.utils.conda import (
    find_conda,
    get_conda_root_prefix,
    get_list_conda_envs,
    get_list_conda_envs_cache,
    get_spyder_conda_channel,
)

if not is_anaconda():
    pytest.skip("Requires conda to be installed", allow_module_level=True)

if os.name == 'nt':
    TEST_PYEXEC = 'c:/miniconda/envs/foobar/python.exe'
else:
    TEST_PYEXEC = '/miniconda/envs/foobar/bin/python'


def test_get_conda_root_prefix():
    output = get_conda_root_prefix(TEST_PYEXEC)
    if os.name == 'nt':
        assert output == 'c:/miniconda'
    else:
        assert output == '/miniconda'

    assert 'envs' not in get_conda_root_prefix(sys.executable)


@pytest.mark.skipif(not running_in_ci(), reason="Only meant for CIs")
def test_find_conda():
    # Standard test
    assert find_conda()

    # Test with test environment
    pyexec = get_conda_test_env()[1]

    # Temporarily remove CONDA_EXE and MAMBA_EXE, if present
    conda_exe = os.environ.pop('CONDA_EXE', None)
    mamba_exe = os.environ.pop('MAMBA_EXE', None)

    assert find_conda(pyexec)

    # Restore os.environ
    if conda_exe is not None:
        os.environ['CONDA_EXE'] = conda_exe
    if mamba_exe is not None:
        os.environ['MAMBA_EXE'] = mamba_exe


@pytest.mark.skipif(not running_in_ci(), reason="Only meant for CIs")
def test_get_list_conda_envs():
    output = get_list_conda_envs()

    if sys.platform == 'darwin':
        expected_envs = [
            'base',
            'jedi-test-env',
            'spytest-ž',
            'micromamba/envs/test',
            'miniconda3/envs/test'
        ]
    else:
        expected_envs = ['base', 'jedi-test-env', 'spytest-ž', 'test']

    expected_envs = ['Conda: ' + env for env in expected_envs]

    assert set(expected_envs) == set(output.keys())


@pytest.mark.skipif(not running_in_ci(), reason="Only meant for CIs")
def test_get_list_conda_envs_cache():
    time0 = time.time()
    output = get_list_conda_envs_cache()
    time1 = time.time()

    assert output != {}
    assert (time1 - time0) < 0.01


@pytest.mark.skipif(not running_in_ci(), reason="Only meant for CIs")
def test_get_spyder_conda_channel():
    channel, channel_url = get_spyder_conda_channel()
    assert channel == "pypi"
    assert channel_url == "https://conda.anaconda.org/pypi"


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""Tests for the Existing Kernel Connection widget."""

# Third party imports
import sys
import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialogButtonBox, QMessageBox
from unittest.mock import patch

# Local imports
from spyder.config.base import running_in_ci
from spyder.plugins.ipythonconsole.widgets.kernelconnect import (
    KernelConnectionDialog)
from spyder.config.manager import CONF


# =============================================================================
# Fixtures
# =============================================================================
@pytest.fixture
def connection_dialog_factory(qtbot, request):
    """Set up kernel connection dialog."""
    class DialogFactory(object):
        def get_default_dialog(self):
            dialog = KernelConnectionDialog()
            request.addfinalizer(dialog.close)
            return dialog

        def submit_filled_dialog(self, use_keyfile, save_settings):
            dlg = self.get_default_dialog()

            # fill out cf path
            dlg.cf.clear()
            qtbot.keyClicks(dlg.cf, pytest.cf_path)
            # check the remote kernel box
            dlg.rm_group.setChecked(True)
            # fill out hostname
            dlg.hn.clear()
            qtbot.keyClicks(dlg.hn, pytest.hn)
            # fill out port
            dlg.pn.clear()
            qtbot.keyClicks(dlg.pn, str(pytest.pn))
            # fill out username
            dlg.un.clear()
            qtbot.keyClicks(dlg.un, pytest.un)

            if use_keyfile:
                # select ssh keyfile radio
                dlg.kf_radio.setChecked(True)
                assert dlg.kf.isEnabled()

                # fill out ssh keyfile
                dlg.kf.clear()
                qtbot.keyClicks(dlg.kf, pytest.kf)

                # fill out passphrase
                dlg.kfp.clear()
                qtbot.keyClicks(dlg.kfp, pytest.kfp)
            else:
                # select password radio
                dlg.pw_radio.setChecked(True)
                assert dlg.pw.isEnabled()

                # fill out password
                dlg.pw.clear()
                qtbot.keyClicks(dlg.pw, pytest.pw)

            # check save connection settings
            dlg.save_layout.setChecked(save_settings)

            return dlg

    def teardown():
        """Clear existing-kernel config and keyring passwords."""
        CONF.remove_section("existing-kernel")

        try:
            import keyring
            keyring.set_password("spyder_remote_kernel",
                                 "ssh_key_passphrase", "")
            keyring.set_password("spyder_remote_kernel",
                                 "ssh_password", "")
        except Exception:
            pass

    # test form values
    pytest.cf_path = "cf_path"
    pytest.un = "test_username"
    pytest.hn = "test_hostname"
    pytest.pn = 123
    pytest.kf = "test_kf"
    pytest.kfp = "test_kfp"
    pytest.pw = "test_pw"

    request.addfinalizer(teardown)
    return DialogFactory()


# =============================================================================
# Tests
# =============================================================================
def test_connection_dialog_remembers_input_with_ssh_passphrase(
        qtbot, connection_dialog_factory):
    """
    Test that the dialog remembers the user's kernel connection
    settings and ssh key passphrase when the user checks the
    save checkbox.
    """

    dlg = connection_dialog_factory.submit_filled_dialog(use_keyfile=True,
                                                         save_settings=True)

    with patch.object(QMessageBox, 'critical', return_value=None):
        # Press ok and save connection settings
        qtbot.mouseClick(dlg.accept_btns.button(QDialogButtonBox.Ok),
                         Qt.LeftButton)

    # create new dialog and check fields
    new_dlg = connection_dialog_factory.get_default_dialog()
    assert new_dlg.cf.text() == pytest.cf_path
    assert new_dlg.rm_group.isChecked()
    assert new_dlg.hn.text() == pytest.hn
    assert new_dlg.un.text() == pytest.un
    assert new_dlg.pn.text() == str(pytest.pn)
    assert new_dlg.kf.text() == pytest.kf
    if not running_in_ci():
        assert new_dlg.kfp.text() == pytest.kfp


def test_connection_dialog_doesnt_remember_input_with_ssh_passphrase(
        qtbot, connection_dialog_factory):
    """
    Test that the dialog doesn't remember the user's kernel
    connection settings and ssh key passphrase when the user doesn't
    check the save checkbox.
    """

    dlg = connection_dialog_factory.submit_filled_dialog(use_keyfile=True,
                                                         save_settings=False)

    with patch.object(QMessageBox, 'critical', return_value=None):
        # Press ok and save connection settings
        qtbot.mouseClick(dlg.accept_btns.button(QDialogButtonBox.Ok),
                         Qt.LeftButton)

    # create new dialog and check fields
    new_dlg = connection_dialog_factory.get_default_dialog()
    assert new_dlg.cf.text() == ""
    assert not new_dlg.rm_group.isChecked()
    assert new_dlg.hn.text() == ""
    assert new_dlg.un.text() == ""
    assert new_dlg.pn.text() == "22"
    assert new_dlg.kf.text() == ""
    if not running_in_ci():
        assert new_dlg.kfp.text() == ""


def test_connection_dialog_remembers_input_with_password(
        qtbot, connection_dialog_factory):
    """
    Test that the dialog remembers the user's kernel connection
    settings and ssh password when the user checks the save checkbox.
    """

    dlg = connection_dialog_factory.submit_filled_dialog(use_keyfile=False,
                                                         save_settings=True)

    with patch.object(QMessageBox, 'critical', return_value=None):
        # Press ok and save connection settings
        qtbot.mouseClick(dlg.accept_btns.button(QDialogButtonBox.Ok),
                         Qt.LeftButton)

    # create new dialog and check fields
    new_dlg = connection_dialog_factory.get_default_dialog()
    assert new_dlg.cf.text() == pytest.cf_path
    assert new_dlg.rm_group.isChecked()
    assert new_dlg.hn.text() == pytest.hn
    assert new_dlg.un.text() == pytest.un
    assert new_dlg.pn.text() == str(pytest.pn)
    if not running_in_ci():
        assert new_dlg.pw.text() == pytest.pw


def test_connection_dialog_doesnt_remember_input_with_password(
        qtbot, connection_dialog_factory):
    """
    Test that the dialog doesn't remember the user's kernel
    connection settings and ssh password when the user doesn't
    check the save checkbox.
    """

    dlg = connection_dialog_factory.submit_filled_dialog(use_keyfile=False,
                                                         save_settings=False)

    with patch.object(QMessageBox, 'critical', return_value=None):
        # Press ok and save connection settings
        qtbot.mouseClick(dlg.accept_btns.button(QDialogButtonBox.Ok),
                         Qt.LeftButton)

    # create new dialog and check fields
    new_dlg = connection_dialog_factory.get_default_dialog()
    assert new_dlg.cf.text() == ""
    assert not new_dlg.rm_group.isChecked()
    assert new_dlg.hn.text() == ""
    assert new_dlg.un.text() == ""
    assert new_dlg.pn.text() == "22"
    if not running_in_ci():
        assert new_dlg.pw.text() == ""


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#
"""Tests for close brackets."""

# Third party imports
import pytest
from qtpy.QtGui import QTextCursor

# Local imports
from spyder.utils.qthelpers import qapplication
from spyder.plugins.editor.widgets.codeeditor import CodeEditor
from spyder.plugins.editor.utils.editor import TextHelper
from spyder.plugins.editor.extensions.closebrackets import (
        CloseBracketsExtension)


# =============================================================================
# ---- Fixtures
# =============================================================================
@pytest.fixture
def editor_close_brackets():
    """Set up Editor with close brackets activated."""
    app = qapplication()
    editor = CodeEditor(parent=None)
    kwargs = {}
    kwargs['language'] = 'Python'
    kwargs['close_parentheses'] = True
    editor.setup_editor(**kwargs)
    return editor


# =============================================================================
# ---- Tests
# =============================================================================
def test_bracket_closing_new_line(qtbot, editor_close_brackets):
    """
    Test bracket completion with existing brackets in a new line.

    For spyder-ide/spyder#11217
    """
    editor = editor_close_brackets
    editor.textCursor().insertText('foo(\nbar)')

    # Move from 'foo(\nbar)|' to 'foo(\nbar|)'
    editor.move_cursor(-1)

    qtbot.keyClicks(editor, '()')
    assert editor.toPlainText() == 'foo(\nbar())'
    assert editor.textCursor().columnNumber() == 5
    qtbot.keyClicks(editor, ')')
    assert editor.toPlainText() == 'foo(\nbar())'
    assert editor.textCursor().columnNumber() == 6


@pytest.mark.parametrize(
    'text, expected_text, cursor_column',
    [
        ("(", "()", 1),  # Close brackets
        ("{", "{}", 1),
        ("[", "[]", 1),
    ])
def test_close_brackets(qtbot, editor_close_brackets, text, expected_text,
                        cursor_column):
    """Test insertion of brackets."""
    editor = editor_close_brackets

    qtbot.keyClicks(editor, text)
    assert editor.toPlainText() == expected_text

    assert cursor_column == TextHelper(editor).current_column_nbr()


@pytest.mark.parametrize(
    'text, expected_text, cursor_column',
    [
        ('()', '(())', 2),  # Complete in brackets
        ('{}', '{()}', 2),
        ('[]', '[()]', 2),
        (',', '(),', 1),  # Complete before commas, colons and semi-colons
        (':', '():', 1),
        (';', '();', 1),
    ])
def test_nested_brackets(qtbot, editor_close_brackets, text, expected_text,
                      cursor_column):
    """
    Test completion of brackets inside brackets and before commas,
    colons and semi-colons.
    """
    editor = editor_close_brackets

    qtbot.keyClicks(editor, text)
    editor.move_cursor(-1)
    qtbot.keyClicks(editor, '(')
    assert editor.toPlainText() == expected_text

    assert cursor_column == TextHelper(editor).current_column_nbr()


def test_selected_text(qtbot, editor_close_brackets):
    """Test insert surronding brackets to selected text."""
    editor = editor_close_brackets
    editor.set_text("some text")

    # select some
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, 4)
    editor.setTextCursor(cursor)

    qtbot.keyClicks(editor, "(")
    assert editor.toPlainText() == "(some) text"

    qtbot.keyClicks(editor, "}")
    assert editor.toPlainText() == "({some}) text"

    qtbot.keyClicks(editor, "[")
    assert editor.toPlainText() == "({[some]}) text"


def test_selected_text_multiple_lines(qtbot, editor_close_brackets):
    """Test insert surronding brackets to multiple lines selected text."""
    editor = editor_close_brackets
    text = ("some text\n"
            "\n"
            "some text")
    editor.set_text(text)

    # select until second some
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, 4)
    cursor.movePosition(QTextCursor.Down, QTextCursor.KeepAnchor, 2)
    editor.setTextCursor(cursor)

    qtbot.keyClicks(editor, ")")
    assert editor.toPlainText() == ("(some text\n"
                                    "\n"
                                    "some) text")

    qtbot.keyClicks(editor, "{")
    assert editor.toPlainText() == ("({some text\n"
                                    "\n"
                                    "some}) text")

    qtbot.keyClicks(editor, "]")
    assert editor.toPlainText() == ("({[some text\n"
                                    "\n"
                                    "some]}) text")


def test_complex_completion(qtbot, editor_close_brackets):
    """Test bracket completion in nested brackets."""
    editor = editor_close_brackets
    # Test completion when following character is a right bracket
    editor.textCursor().insertText('foo(bar)')
    editor.move_cursor(-1)
    qtbot.keyClicks(editor, '(')
    assert editor.toPlainText() == 'foo(bar())'
    assert editor.textCursor().columnNumber() == 8
    # Test normal insertion when next character is not a right bracket
    editor.move_cursor(-1)
    qtbot.keyClicks(editor, '[')
    assert editor.toPlainText() == 'foo(bar[())'
    assert editor.textCursor().columnNumber() == 8
    # Test completion when following character is a comma
    qtbot.keyClicks(editor, ',')
    editor.move_cursor(-1)
    qtbot.keyClicks(editor, '{')
    assert editor.toPlainText() == 'foo(bar[{},())'
    assert editor.textCursor().columnNumber() == 9


def test_bracket_closing(qtbot, editor_close_brackets):
    """Test bracket completion with existing brackets."""
    editor = editor_close_brackets
    editor.textCursor().insertText('foo(bar(x')
    qtbot.keyClicks(editor, ')')
    assert editor.toPlainText() == 'foo(bar(x)'
    assert editor.textCursor().columnNumber() == 10
    qtbot.keyClicks(editor, ')')
    assert editor.toPlainText() == 'foo(bar(x))'
    assert editor.textCursor().columnNumber() == 11
    # same ')' closing with existing brackets starting at 'foo(bar(x|))'
    editor.move_cursor(-2)
    qtbot.keyClicks(editor, ')')
    assert editor.toPlainText() == 'foo(bar(x))'
    assert editor.textCursor().columnNumber() == 10
    qtbot.keyClicks(editor, ')')
    assert editor.toPlainText() == 'foo(bar(x))'
    assert editor.textCursor().columnNumber() == 11


def test_activate_deactivate(qtbot, editor_close_brackets):
    """Test activating/desctivating close quotes editor extension."""
    editor = editor_close_brackets
    bracket_extension = editor.editor_extensions.get(CloseBracketsExtension)

    qtbot.keyClicks(editor, "(")
    assert editor.toPlainText() == "()"

    editor.set_text("")
    bracket_extension.enabled = False
    qtbot.keyClicks(editor, "(")
    assert editor.toPlainText() == "("

    editor.set_text("")
    bracket_extension.enabled = True
    qtbot.keyClicks(editor, "(")
    assert editor.toPlainText() == "()"


if __name__ == '__main__':
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Status bar widget that shows content that comes from the IPython console.
"""

# Standard library imports
import functools

# Local imports
from spyder.api.shellconnect.mixins import ShellConnectMixin
from spyder.api.widgets.status import StatusBarWidget


class ShellConnectStatusBarWidget(StatusBarWidget, ShellConnectMixin):
    """
    Base class for status bar widgets whose info depends on the current shell.
    """

    def __init__(self, parent):
        super().__init__(parent)

        # Mapping from shellwidgets to their corresponding statuses
        self.shellwidget_to_status = {}

        # Current associated shellwidget
        self.current_shellwidget = None

    # ---- Public API
    # -------------------------------------------------------------------------
    def update_status(self, status):
        """
        Actions to take to update the widget status after switching consoles.
        """
        raise NotImplementedError

    def on_kernel_start(self, shellwidget):
        """Actions to take when the kernel starts."""
        raise NotImplementedError

    def config_spyder_kernel(self, shellwidget):
        """Actions to take to configure the kernel."""
        pass

    # ---- ShellConnectMixin API
    # -------------------------------------------------------------------------
    def set_shellwidget(self, shellwidget):
        """
        Actions to take when setting (i.e. giving focus to) a shellwidget.
        """
        self.current_shellwidget = None
        if shellwidget in self.shellwidget_to_status:
            status = self.shellwidget_to_status[shellwidget]
            if status is None:
                # This means the console failed to start or there was a problem
                # obtaining the status from the kernel. So, we don't need to
                # show this widget in those cases.
                self.hide()
            else:
                self.show()
                self.current_shellwidget = shellwidget
                self.update_status(status)

    def add_shellwidget(self, shellwidget):
        """Actions to take when adding a shellwidget."""
        shellwidget.sig_config_spyder_kernel.connect(
            functools.partial(self.config_spyder_kernel, shellwidget)
        )
        shellwidget.kernel_handler.sig_kernel_is_ready.connect(
            # We can't use functools.partial here because it gives memory leaks
            # in Python versions older than 3.10
            lambda sw=shellwidget: self.on_kernel_start(sw)
        )

        self.shellwidget_to_status[shellwidget] = None
        self.set_shellwidget(shellwidget)

    def remove_shellwidget(self, shellwidget):
        """Actions to take when removing a shellwidget."""
        if shellwidget in self.shellwidget_to_status:
            self.shellwidget_to_status.pop(shellwidget)

    def add_errored_shellwidget(self, shellwidget):
        """Actions to take when a shellwidget errored."""
        self.shellwidget_to_status[shellwidget] = None
        self.set_shellwidget(shellwidget)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Update Manager Plugin.
"""

# Local imports
from spyder.api.plugins import Plugins, SpyderPluginV2
from spyder.api.translations import _
from spyder.api.plugin_registration.decorators import (
    on_plugin_available,
    on_plugin_teardown
)
from spyder.plugins.updatemanager.container import (
    UpdateManagerActions,
    UpdateManagerContainer
)
from spyder.plugins.mainmenu.api import ApplicationMenus, HelpMenuSections


class UpdateManager(SpyderPluginV2):
    NAME = 'update_manager'
    REQUIRES = [Plugins.Preferences]
    OPTIONAL = [Plugins.MainMenu, Plugins.StatusBar]
    CONTAINER_CLASS = UpdateManagerContainer
    CONF_SECTION = 'update_manager'
    CONF_FILE = False
    CAN_BE_DISABLED = False

    # ---- SpyderPluginV2 API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _('Update Manager')

    @classmethod
    def get_icon(cls):
        return cls.create_icon('genprefs')

    @staticmethod
    def get_description():
        return _('Manage application updates.')

    # ---- Plugin initialization
    def on_initialize(self):
        pass

    @on_plugin_available(plugin=Plugins.Preferences)
    def on_preferences_available(self):
        # Register conf page
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(self)

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_main_menu_available(self):
        if self.is_plugin_enabled(Plugins.Shortcuts):
            if self.is_plugin_available(Plugins.Shortcuts):
                self._populate_help_menu()
        else:
            self._populate_help_menu()

    @on_plugin_available(plugin=Plugins.StatusBar)
    def on_statusbar_available(self):
        # Add status widget
        statusbar = self.get_plugin(Plugins.StatusBar)
        statusbar.add_status_widget(self.update_manager_status)

    # ---- Plugin teardown
    @on_plugin_teardown(plugin=Plugins.StatusBar)
    def on_statusbar_teardown(self):
        # Remove status widget if created
        statusbar = self.get_plugin(Plugins.StatusBar)
        statusbar.remove_status_widget(self.update_manager_status.ID)

    @on_plugin_teardown(plugin=Plugins.Preferences)
    def on_preferences_teardown(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.deregister_plugin_preferences(self)

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_main_menu_teardown(self):
        self._depopulate_help_menu()

    def on_close(self, _unused=True):
        # The container is closed directly in the plugin registry
        pass

    def on_mainwindow_visible(self):
        """Actions after the mainwindow in visible."""
        container = self.get_container()

        # Initialize status.
        # Note that NO_STATUS also hides the statusbar widget.
        container.update_manager_status.set_no_status()

        # Check for updates on startup
        if self.get_conf('check_updates_on_startup'):
            container.start_check_update(startup=True)

    # ---- Private API
    # ------------------------------------------------------------------------
    def _populate_help_menu(self):
        """Add update action and menu to the Help menu."""
        mainmenu = self.get_plugin(Plugins.MainMenu)
        mainmenu.add_item_to_application_menu(
            self.check_update_action,
            menu_id=ApplicationMenus.Help,
            section=HelpMenuSections.Support,
            before_section=HelpMenuSections.ExternalDocumentation)

    @property
    def _window(self):
        return self.main.window()

    def _depopulate_help_menu(self):
        """Remove update action from the Help main menu."""
        mainmenu = self.get_plugin(Plugins.MainMenu)
        mainmenu.remove_item_from_application_menu(
            UpdateManagerActions.SpyderCheckUpdateAction,
            menu_id=ApplicationMenus.Help)

    # ---- Public API
    # ------------------------------------------------------------------------
    @property
    def check_update_action(self):
        """Check if a new version of Spyder is available."""
        return self.get_container().check_update_action

    @property
    def update_manager_status(self):
        """Get Update manager statusbar widget"""
        return self.get_container().update_manager_status



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for the autoindent features
"""

# Third party imports
from qtpy.QtGui import QTextCursor
from qtpy.QtWidgets import QApplication
import pytest

# Local imports
from spyder.utils.qthelpers import qapplication
from spyder.py3compat import to_text_string
from spyder.plugins.editor.widgets.codeeditor import CodeEditor


# ---- Fixtures
def get_indent_fix(text, indent_chars=" " * 4, tab_stop_width_spaces=4,
                   sol=False, forward=True, language='Python'):
    """Return text with last line's indentation fixed."""
    app = qapplication()
    editor = CodeEditor(parent=None)
    editor.setup_editor(language=language, indent_chars=indent_chars,
                        tab_stop_width_spaces=tab_stop_width_spaces)

    editor.set_text(text)
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.End)
    if sol:
        lines = text.splitlines(True)
        repeat = len(lines[-1].lstrip())
        cursor.movePosition(QTextCursor.Left, n=repeat)
    editor.setTextCursor(cursor)
    editor.fix_indent(forward=forward)
    return to_text_string(editor.toPlainText())


# ---- Tests
def test_simple_tuple():
    text = get_indent_fix("this_tuple = (1, 2)\n")
    assert text == "this_tuple = (1, 2)\n"


def test_def_with_newline():
    text = get_indent_fix("\ndef function():\n")
    assert text == "\ndef function():\n    ", repr(text)


def test_def_with_indented_comment():
    text = get_indent_fix("def function():\n    # Comment\n")
    assert text == "def function():\n    # Comment\n    ", repr(text)


def test_brackets_alone():
    text = get_indent_fix("def function():\n    print []\n")
    assert text == "def function():\n    print []\n    ", repr(text)


def test_simple_def():
    text = get_indent_fix("def function():\n")
    assert text == "def function():\n    ", repr(text)


def test_open_parenthesis():
    # An open parenthesis with no item is followed by a hanging indent
    text = get_indent_fix("open_parenthesis(\n")
    assert text == "open_parenthesis(\n    ", repr(text)


def test_open_bracket():
    # An open bracket with no item is followed by a hanging indent
    text = get_indent_fix("open_bracket[\n")
    assert text == "open_bracket[\n    ", repr(text)


def test_open_curly():
    # An open curly bracket with no item is followed by a hanging indent
    text = get_indent_fix("open_curly{\n")
    assert text == "open_curly{\n    ", repr(text)


def test_align_on_parenthesis():
    # An open parenthesis with one or more item is followed by an indent
    # up to the parenthesis.
    text = get_indent_fix("parenthesis_w_item = (1,\n")
    assert text == "parenthesis_w_item = (1,\n                      ", repr(text)


def test_align_on_bracket():
    # An open bracket with one or more item is followed by an indent
    # up to the parenthesis.
    text = get_indent_fix("bracket_w_item = [1,\n")
    assert text == "bracket_w_item = [1,\n                  ", repr(text)


def test_align_on_curly():
    # An open curly bracket with one or more item is followed by an indent
    # up to the parenthesis.
    text = get_indent_fix("curly_w_item = {1,\n")
    assert text == "curly_w_item = {1,\n                ", repr(text)


def test_keep_unindent():
    # Keep line unindented if there is more than one line under the statement
    text = ("    def foo(bar):\n"
            "        generic = bar\n"
            "    \n"
            "    keep_unindent\n")
    correct_text = ("    def foo(bar):\n"
                    "        generic = bar\n"
                    "    \n"
                    "    keep_unindent\n")
    text = get_indent_fix(text, sol=True)
    assert text == correct_text, repr(text)


def test_keep_unindent_fix_indent():
    # Keep line unindented but fix indent if not multiple of len(indent_chars)
    text = ("    for x in range(n):\n"
            "        increment += 1\n"
            "  \n"
            "  keep_unindent\n")
    correct_text = ("    for x in range(n):\n"
                    "        increment += 1\n"
                    "  \n"
                    "    keep_unindent\n")
    text = get_indent_fix(text, sol=True)
    assert text == correct_text, repr(text)


def test_keep_unindent_if_blank():
    # Keep line unindented if return is pressed on a line which is both
    # blank and unindented.
    text = ("    def f(x):\n"
            "        return x\n"
            "\n"
            "")
    text = get_indent_fix(text)
    assert text == "    def f(x):\n        return x\n\n", repr(text)


def test_first_line():
    # Test fix_indent() when the cursor is on the first line.
    text = get_indent_fix("import numpy")
    assert text == "import numpy", repr(text)


def test_indentation_with_spaces():
    text_input_expected_comment = [
        ("tags = ['(a)', '(b)', '(c)']\n", "tags = ['(a)', '(b)', '(c)']\n",
         "test_commented_brackets"),
        ("s = a[(a['a'] == l) & (a['a'] == 1)]['a']\n",
         "s = a[(a['a'] == l) & (a['a'] == 1)]['a']\n",
         "test_balanced_brackets"),
        ("a = (a  #  some comment\n", "a = (a  #  some comment\n     ",
         "test_inline_comment"),
        ("len(a) == 1\n", "len(a) == 1\n",
         "test_balanced_brackets_not_ending_in_bracket"),
        ("x = f(\n", "x = f(\n      ",
         "test_short_open_bracket_not_hanging_indent"),
        ("def some_func():\n    return 10\n",
         "def some_func():\n    return 10\n",
         "test_return"),
        ("def some_func():\n    returns = 10\n",
         "def some_func():\n    returns = 10\n    ",
         "test_return_not_keyword"),
        ("foo = 1  # Comment open parenthesis (\n",
         "foo = 1  # Comment open parenthesis (\n",
         "test_comment_with parenthesis")
        ]
    for text_input, expected, comment in text_input_expected_comment:
        text = get_indent_fix(text_input)
        assert text == expected, comment


def test_def_with_unindented_comment():
    text = get_indent_fix("def function():\n# Comment\n")
    assert text == "def function():\n# Comment\n", repr(text)


def test_issue_5606():
    # See spyder-ide/spyder#5606
    text_input_expected_comment = [
        ("if True:\n    print(str(str('foo' \n+ str('bar'))))",
         "if True:\n    print(str(str('foo' \n                  + str('bar'))))",
         "Before + 1"),
        ("if True:\n    print(str(str('foo' +\n str('bar'))))",
         "if True:\n    print(str(str('foo' +\n                  str('bar'))))",
         "After + 1"),
        ("if True:\n    print(str(str(str('foo') \n+ 'bar')))",
         "if True:\n    print(str(str(str('foo') \n                  + 'bar')))",
         "Before + 2"),
        ("if True:\n    print(str(str(str('foo') +\n 'bar')))",
         "if True:\n    print(str(str(str('foo') +\n                  'bar')))",
         "After + 2"),
        ("if True:\n    some_long_variable_name = (\n'foo' + 'bar')",
         "if True:\n    some_long_variable_name = (\n        'foo' + 'bar')",
         "Open paren case"),
        ("if True:\n    (print(str(('Long string ' + 'Yeah')\n.format(some_long_variable_name))",
         "if True:\n    (print(str(('Long string ' + 'Yeah')\n               .format(some_long_variable_name))",
         "Before . 1"),
        ("if True:\n    (print(str(('Long string ' + 'Yeah').\nformat(some_long_variable_name))",
         "if True:\n    (print(str(('Long string ' + 'Yeah').\n               format(some_long_variable_name))",
         "After . 1"),
        ]
    for text_input, expected, comment in text_input_expected_comment:
        text = get_indent_fix(text_input)
        assert text == expected, comment


def test_issue_7377():
    # See spyder-ide/spyder#7377
    text_input_expected_comment = [
        ("print('#1 w(Earth) =', w)\n",
         "print('#1 w(Earth) =', w)\n",
         "Indentation made"),
        ("print('1 w(Earth) =', w)\n",
         "print('1 w(Earth) =', w)\n",
         "Indentation made"),
        ]
    for text_input, expected, comment in text_input_expected_comment:
        text = get_indent_fix(text_input)
        assert text == expected, comment


def test_issue_887():
    # See spyder-ide/spyder#887
    text_input_expected_comment = [
        ("df = pd.read_csv(fpath,\n" +
         "                 # HINT indexing begins with 0\n" +
         "                 header=0, # headers 1st row\n" +
         "                 index_col=0, #timestamps 1st column\n" +
         "                 # specify as seconds to recognize as epoch timestamp...\n" +
         "                 # it's local time (i.e...)\n" +
         "                 # (section 9.10.3, http://......)\n" +
         "                 date_parser=lambda x: pd.to_datetime(x, units='s'),\n" +
         "            parse_dates=True)",
         "df = pd.read_csv(fpath,\n" +
         "                 # HINT indexing begins with 0\n" +
         "                 header=0, # headers 1st row\n" +
         "                 index_col=0, #timestamps 1st column\n" +
         "                 # specify as seconds to recognize as epoch timestamp...\n" +
         "                 # it's local time (i.e...)\n" +
         "                 # (section 9.10.3, http://......)\n" +
         "                 date_parser=lambda x: pd.to_datetime(x, units='s'),\n" +
         "                 parse_dates=True)",
         "Indentation made"),
        ("df = pd.read_csv(fpath,\n" +
         "                 # HINT indexing begins with 0\n" +
         "                 header=0, # headers 1st row\n" +
         "                 index_col=0, #timestamps 1st column\n" +
         "                 # specify as seconds to recognize as epoch timestamp...\n" +
         "                 # it's local time (i.e...)\n" +
         "                 # (section 9.10.3, http://......)\n" +
         "                 #date_parser=lambda x: pd.to_datetime(x, units='s'),\n" +
         "            parse_dates=True)",
         "df = pd.read_csv(fpath,\n" +
         "                 # HINT indexing begins with 0\n" +
         "                 header=0, # headers 1st row\n" +
         "                 index_col=0, #timestamps 1st column\n" +
         "                 # specify as seconds to recognize as epoch timestamp...\n" +
         "                 # it's local time (i.e...)\n" +
         "                 # (section 9.10.3, http://......)\n" +
         "                 #date_parser=lambda x: pd.to_datetime(x, units='s'),\n" +
         "                 parse_dates=True)",
         "Indentation made"),        ]
    for text_input, expected, comment in text_input_expected_comment:
        text = get_indent_fix(text_input)
        assert text == expected, comment


def test_keywords():
    # 'def', 'for', 'if', 'while', 'with', 'class', 'elif', 'except'
    text_input_expected_comment = [
        ("def func():\n", "def func():\n    ",
         "function declaration"),
        ("def function(\n", "def function(\n        ",
         "partial function declaration"),
        ("def f(\n", "def f(\n        ",
         "partial function declaration with short name"),
        ("def func(): return\n", "def func(): return\n", "single line return"),
        ("def func():\n    if True:\n        return\n",
         "def func():\n    if True:\n        return\n    ",
         "return in if block"),
        ("if True:\n    varname = multipass\n",
         "if True:\n    varname = multipass\n    ",
         "endswith pass, but not keyword"),
        ("def func():\n    return()\n", "def func():\n    return()\n",
         "no space between return and bracket"),
        ("for i in range(\n", "for i in range(\n        ", "partial for"),
        ]
    for text_input, expected, comment in text_input_expected_comment:
        text = get_indent_fix(text_input)
        assert text == expected, comment


def test_hanging_indentation_without_block():
    text_input_expected_comment = [
        ("function(\n",
         "function(\n    ",
         "function call"),
        ("f(\n",
         "f(\n  ",
         "function call with short name"),
        ("def f():\n    return \\\n", "def f():\n    return \\\n        ",
         "line cont. return"),
        ("def f():\n    return lambda: \\\n",
         "def f():\n    return lambda: \\\n        ",
         "return but indent")
        ]
    for text_input, expected, comment in text_input_expected_comment:
        text = get_indent_fix(text_input)
        assert text == expected, comment


def test_no_hanging_indentation():
    text_input_expected_comment = [
        ("f('(')\n",
         "f('(')\n",
         "open bracket in string"),
        ("f('def f(')",
         "f('def f(')",
         "keyword in string"),
        ("'a(a,'\n",
         "'a(a,'\n",
         "bracket in string"),
        ("def f():\n    return()\n",
         "def f():\n    return()\n",
         "return without space after"),
        ("print('#1 w(Earth) =', w)\n",
         "print('#1 w(Earth) =', w)\n",
         "# in string")
        ]
    for text_input, expected, comment in text_input_expected_comment:
        text = get_indent_fix(text_input)
        assert text == expected, comment


def test_unintent_in_block():
    # return, yield, pass, continue, break, raise
    text = get_indent_fix("def f(x):\n    if True:\n        return\n")
    assert text == "def f(x):\n    if True:\n        return\n    ", repr(text)


def test_multi_line_lists():
#         "corbin_dallas = multipass"
# if True: pass
    # if (
    #     pass
    # if True: if True: (invalid syntax)
    # if True: return
    # def f():\n# comment\n#
    text_input_expected_comment = [
        ('[\n    ("",\n     "",\n     "")\n',
         '[\n    ("",\n     "",\n     "")\n    ',
         "multiline args"),
        ('[\n    ("",\n     "",\n     ""),\n',
         '[\n    ("",\n     "",\n     ""),\n    ',
         "multiline args with comma")
        ]
    for text_input, expected, comment in text_input_expected_comment:
        text = get_indent_fix(text_input)
        assert text == expected, comment


def test_string_literals():
    text_input_expected_comment = [
        ("'''\n", "'''\n", "bare opening triple-quote"),
        ("def f():\n    '''\n", "def f():\n    '''\n    ",
         "opening triple-quote in function declaration"),
        ("'''Docstring'''\n", "'''Docstring'''\n",
         "triple-quoted string literal"),
        ("", "", ""),
        ("", "", ""),
        ]
    for text_input, expected, comment in text_input_expected_comment:
        text = get_indent_fix(text_input)
        assert text == expected, comment


# ---- Tabs tests
def test_indentation_with_tabs():
    text_input_expected_comment = [
        ("this_tuple = (1, 2)\n", "this_tuple = (1, 2)\n", "simple tuple"),
        ("\ndef function():\n", "\ndef function():\n\t", "def with new line"),
        ("def function():\n\t# Comment\n", "def function():\n\t# Comment\n\t",
         "test with indented comment"),
        ("def function():\n\tprint []\n", "def function():\n\tprint []\n\t",
         "test brackets alone"),
        ("\nsome_long_name = {\n", "\nsome_long_name = {\n\t",
         "indentation after opening bracket"),
        ("def function():\n", "def function():\n\t", "test simple def"),
        ("open_parenthesis(\n", "open_parenthesis(\n\t",
         "open parenthesis"),
        ("tags = ['(a)', '(b)', '(c)']\n", "tags = ['(a)', '(b)', '(c)']\n",
         "test_commented_brackets"),
        ("s = a[(a['a'] == l) & (a['a'] == 1)]['a']\n",
         "s = a[(a['a'] == l) & (a['a'] == 1)]['a']\n",
         "test_balanced_brackets"),
        ("def some_func():\n\treturn 10\n", "def some_func():\n\treturn 10\n",
         "test_return"),
        ("def some_func():\n\treturns = 10\n",
         "def some_func():\n\treturns = 10\n\t",
         "test_return_not_keyword"),
        ("def function():\n# Comment\n", "def function():\n# Comment\n",
         "test_def_with_unindented_comment")
        ]

    for text_input, expected, comment in text_input_expected_comment:
        for tab_stop_width_spaces in [1, 2, 3, 4, 5, 6, 7, 8]:
            text = get_indent_fix(text_input, indent_chars="\t",
                                  tab_stop_width_spaces=tab_stop_width_spaces)
            assert text == expected, comment


def test_indentation_with_tabs_parenthesis():
    """Simple parenthesis indentation test with different tab stop widths."""
    text_input_expected_tab = [
        ("print(\n)", "print(\n\t)", 1),
        ("print(\n)", "print(\n\t)", 2),
        ("print(\n)", "print(\n\t\t)", 3),
        ("print(\n)", "print(\n\t  )", 4),
        ("print(\n)", "print(\n\t )", 5),
        ("print(\n)", "print(\n\t)", 6),
        ("print(\n)", "print(\n      )", 7),
        ("print(\n)", "print(\n      )", 8),
        ("a = (a  #  some comment\n", "a = (a  #  some comment\n\t ", 4)
        ]
    for text_input, expected, tab_stop_width_spaces in text_input_expected_tab:
        text = get_indent_fix(text_input, indent_chars="\t",
                              tab_stop_width_spaces=tab_stop_width_spaces)
        assert text == expected, tab_stop_width_spaces


def test_unindentation_with_tabs():
    text_input = "\tx = 1"
    expected_text = "x = 1"
    for tab_stop_width_spaces in [1, 2, 3, 4, 5, 6, 7, 8]:
        text = get_indent_fix(text_input, indent_chars="\t",
                              tab_stop_width_spaces=tab_stop_width_spaces,
                              forward=False)
        assert text == expected_text


# ---- Simple indentation tests
def test_simple_indentation():
    # language None deactivate smart indentation
    text_input_expected_comment = [
        ("hola\n", "hola\n", "witout indentation"),
        ("  hola\n", "  hola\n  ", "some indentation"),
        ("\thola\n", "\thola\n\t", "tab indentation"),
        ("  hola(\n", "  hola(\n  ", "line with parenthesis")
        ]
    for text_input, text_expected, comment in text_input_expected_comment:
        text = get_indent_fix(text_input, language=None)
        assert text == text_expected, comment


def test_copy_paste_autoindent(codeeditor):
    """Test copy pasting text into the editor at different indent."""
    editor = codeeditor
    text = ("if a:\n    b\n    if c:\n        d\n    if e:\n        f\n")
    editor.set_text(text)
    # Copy
    cursor = editor.textCursor()
    cursor.setPosition(30)
    cursor.setPosition(51, QTextCursor.KeepAnchor)
    editor.setTextCursor(cursor)
    editor.copy()

    d = {
        0: "if a:\n    b\nd\nif e:\n    f\n    if c:\n        d\n    if e:\n        f\n",
        4: "if a:\n    b\n    d\nif e:\n    f\n    if c:\n        d\n    if e:\n        f\n",
        8: "if a:\n    b\n        d\n    if e:\n        f\n    if c:\n        d\n    if e:\n        f\n",
        }
    # Paste
    for indent in d.keys():
        editor.set_text(text)
        cursor = editor.textCursor()
        cursor.setPosition(11)
        cursor.insertText("\n" + indent * ' ')
        editor.setTextCursor(cursor)
        editor.paste()
        assert editor.toPlainText() == d[indent]

    # Same thing but with copying the spaces
    editor.set_text(text)
    # Copy
    cursor = editor.textCursor()
    cursor.setPosition(30-4)
    cursor.setPosition(51, QTextCursor.KeepAnchor)
    editor.setTextCursor(cursor)
    editor.copy()

    d = {
        0: "if a:\n    b\n    d\nif e:\n    f\n    if c:\n        d\n    if e:\n        f\n",
        4: "if a:\n    b\n    d\nif e:\n    f\n    if c:\n        d\n    if e:\n        f\n",
        8: "if a:\n    b\n        d\n    if e:\n        f\n    if c:\n        d\n    if e:\n        f\n",
        }
    # Paste
    for indent in d.keys():
        editor.set_text(text)
        cursor = editor.textCursor()
        cursor.setPosition(11)
        cursor.insertText("\n" + indent * ' ')
        editor.setTextCursor(cursor)
        editor.paste()
        assert editor.toPlainText() == d[indent]


if __name__ == "__main__":
    pytest.main()
